"""Run time methods."""
import logging
import os
import shutil
import sys
from subprocess import PIPE, STDOUT, CalledProcessError, Popen

from .util import remove_existing_file


class BatchJob(object):
    """Batch job."""

    def __init__(self, rte, wrapper=""):
        """Construct batch job.

        Args:
            rte (dict): Run time environment.
            wrapper (str, optional): Wrapper around command. Defaults to "".

        """
        self.rte = rte
        self.wrapper = wrapper
        logging.debug("Constructed BatchJob")

    def run(self, cmd):
        """Run command.

        Args:
            cmd (str): Command to run.

        Raises:
            CalledProcessError: Command failed.
            RuntimeError: No command provided!

        """
        if cmd is None:
            raise RuntimeError("No command provided!")
        cmd = self.wrapper + " " + cmd

        if "OMP_NUM_THREADS" in self.rte:
            logging.info("BATCH: %s", self.rte["OMP_NUM_THREADS"])
        logging.info("Batch running %s", cmd)

        process = Popen(  # noqaS602
            cmd,
            shell=True,  # noqaS602
            env=self.rte,
            stdout=PIPE,
            stderr=STDOUT,
            universal_newlines=True,
            bufsize=1,
        )
        # Poll process for new output until finished
        while True:
            nextline = process.stdout.readline()
            if nextline == "" and process.poll() is not None:
                break
            sys.stdout.write(nextline)
            sys.stdout.flush()

        return_code = process.wait()
        if return_code != 0:
            raise CalledProcessError(return_code, cmd)


class SURFEXBinary(object):
    """SURFEX binary class."""

    def __init__(self, binary, batch, iofile, settings, input_data, **kwargs):
        """Surfex binary task.

        Args:
            binary (str): Binary to run
            batch (surfex.Batch): Batch object to run command in
            iofile (surfex.SurfexIO): Input file to command.
            settings (f90nml.Namelist): Fortran namelist namelist
            input_data (surfex.InputDataToSurfexBinaries): Input to binary
            kwargs (dict): Key word arguments.

        Raises:
            FileNotFoundError: Input file not found
            RuntimeError: Execution failed

        """
        self.binary = binary
        self.batch = batch
        self.iofile = iofile
        self.settings = settings

        self.surfout = kwargs.get("surfout")
        self.archive_data = kwargs.get("archive_data")

        self.print_namelist = False
        if "print_namelist" in kwargs:
            self.print_namelist = kwargs["print_namelist"]

        self.pgdfile = kwargs.get("pgdfile")

        # Set input
        self.input_data = input_data
        self.input_data.prepare_input()

        if os.path.exists("OPTIONS.nam"):
            os.remove("OPTIONS.nam")
        self.settings.write("OPTIONS.nam")
        with open("OPTIONS.nam", mode="r", encoding="utf-8") as file_handler:
            content = file_handler.read()
        if self.print_namelist:
            logging.info(content)

        if self.iofile.need_pgd and self.pgdfile is not None:
            logging.debug(self.pgdfile.filename)
            try:
                logging.info("PGD is %s", self.pgdfile.filename)
                if self.pgdfile.input_file is not None and os.path.abspath(
                    self.pgdfile.filename
                ) != os.path.abspath(self.pgdfile.input_file):
                    remove_existing_file(self.pgdfile.input_file, self.pgdfile.filename)
                    os.symlink(self.pgdfile.input_file, self.pgdfile.filename)
                if not os.path.exists(self.pgdfile.filename):
                    raise FileNotFoundError(f"PGD {self.pgdfile.filename} not found!")
            except FileNotFoundError:
                raise FileNotFoundError("Could not set PGD") from FileNotFoundError
            if self.surfout is not None:
                try:
                    logging.info("PREP is %s", self.iofile.filename)
                    if self.iofile.input_file is not None and os.path.abspath(
                        self.iofile.filename
                    ) != os.path.abspath(self.iofile.input_file):
                        remove_existing_file(self.iofile.input_file, self.iofile.filename)
                        os.symlink(self.iofile.input_file, self.iofile.filename)
                    if not os.path.exists(self.iofile.filename):
                        raise FileNotFoundError(f"PREP {self.iofile.filename} not found!")
                except FileNotFoundError:
                    raise FileNotFoundError("Could not set PREP") from FileNotFoundError

        cmd = self.binary
        logging.info("Running %s with settings OPTIONS.nam", cmd)
        try:
            self.batch.run(cmd)
        except Exception as exc:
            raise RuntimeError(repr(exc)) from Exception

        listings = [
            "LISTING_PGD0.txt",
            "LISTING_PREP0.txt",
            "LISTING_OFFLINE0.txt",
            "LISTING_SODA0.txt",
        ]
        for listing in listings:
            if os.path.exists(listing):
                with open(listing, mode="r", encoding="utf-8") as file_handler:
                    content = file_handler.read()
                logging.info("Content of %s:", listing)
                logging.debug(content)

        # Archive output
        self.iofile.archive_output_file()
        if self.surfout is not None:
            self.surfout.archive_output_file()
        if self.archive_data is not None:
            self.archive_data.archive_files()


class PerturbedOffline(SURFEXBinary):
    """Pertubed offline."""

    def __init__(
        self,
        binary,
        batch,
        io,
        pert_number,
        settings,
        input_data,
        surfout=None,
        archive_data=None,
        pgdfile=None,
        print_namelist=False,
        negpert=False,
    ):
        """Perturbed offline.

        Args:
            binary (_type_): _description_
            batch (_type_): _description_
            io (_type_): _description_
            pert_number (_type_): _description_
            settings (_type_): _description_
            input_data (_type_): _description_
            surfout (_type_, optional): _description_. Defaults to None.
            archive_data (_type_, optional): _description_. Defaults to None.
            pgdfile (str, optional): _description_. Defaults to None.
            print_namelist (bool, optional): _description_. Defaults to False.
            negpert (bool, optional): _description_. Defaults to False.

        """
        self.pert_number = pert_number
        settings["nam_io_varassim"]["LPRT"] = True
        settings["nam_var"]["nivar"] = int(pert_number)
        # Handle negative pertubations
        if negpert:
            nvar = int(settings["nam_var"]["nvar"])
            ipert = 0
            npert = 1
            for nvi in range(0, nvar):
                key = "nncv(" + str(nvi + 1) + ")"
                val = int(settings["nam_var"][key])
                # Check if active
                if val == 1:
                    npert = 1
                else:
                    npert = npert + 1
                for __ in range(0, npert):
                    ipert = ipert + 1
                    key = "xtprt_m(" + str(ipert) + ")"
                    val = settings["nam_var"][key]
                    settings["nam_var"][key] = -val
        SURFEXBinary.__init__(
            self,
            binary,
            batch,
            io,
            settings,
            input_data,
            surfout=surfout,
            archive_data=archive_data,
            pgdfile=pgdfile,
            print_namelist=print_namelist,
        )


class Masterodb(object):
    """Masterodb."""

    def __init__(
        self,
        pgdfile,
        prepfile,
        surffile,
        settings,
        input_data,
        binary=None,
        archive_data=None,
        print_namelist=True,
        batch=None,
    ):
        """Masterodb.

        Args:
            pgdfile (_type_): _description_
            prepfile (_type_): _description_
            surffile (_type_): _description_
            settings (_type_): _description_
            input_data (_type_): _description_
            binary (_type_, optional): _description_. Defaults to None.
            archive_data (_type_, optional): _description_. Defaults to None.
            print_namelist (bool, optional): _description_. Defaults to True.
            batch (_type_, optional): _description_. Defaults to None.

        Raises:
            FileNotFoundError: _description_
            FileNotFoundError: _description_

        """
        self.settings = settings
        self.binary = binary
        self.prepfile = prepfile
        self.surfout = surffile
        self.batch = batch
        self.pgdfile = pgdfile
        self.input = input_data
        self.archive = archive_data
        self.print_namelist = print_namelist

        # Set input
        self.input.prepare_input()

        # Prepare namelist
        if os.path.exists("EXSEG1.nam"):
            os.remove("EXSEG1.nam")

        self.settings.write("EXSEG1.nam")
        with open("EXSEG1.nam", mode="r", encoding="utf-8") as file_handler:
            content = file_handler.read()
        if self.print_namelist:
            logging.info(content)

        logging.info("PGD file for MASTERODB %s", self.pgdfile.filename)
        if self.pgdfile.input_file is not None and os.path.abspath(
            self.pgdfile.filename
        ) != os.path.abspath(self.pgdfile.input_file):
            logging.info("Input PGD file is: %s", self.pgdfile.input_file)
            remove_existing_file(self.pgdfile.input_file, self.pgdfile.filename)
            os.symlink(self.pgdfile.input_file, self.pgdfile.filename)

        if not os.path.exists(self.pgdfile.filename):
            logging.error("PGD not found! %s", self.pgdfile.filename)
            raise FileNotFoundError(self.pgdfile.filename)

        logging.info("PREP file for MASTERODB %s", self.prepfile.filename)
        if self.prepfile.input_file is not None and os.path.abspath(
            self.prepfile.filename
        ) != os.path.abspath(self.prepfile.input_file):

            logging.info("Input PREP file is: %s", self.prepfile.input_file)
            remove_existing_file(self.prepfile.input_file, self.prepfile.filename)
            os.symlink(self.prepfile.input_file, self.prepfile.filename)

        if not os.path.exists(self.prepfile.filename):
            logging.error("PREP not found! %s", self.prepfile.filename)
            raise FileNotFoundError(self.prepfile.filename)

        # Archive if we have run the binary
        if self.binary is not None:
            logging.info("Run binary with namelist EXSEG1.nam: %s", self.binary)
            self.batch.run(self.binary)

    def archive_output(self):
        """Archive output."""
        # Archive output
        self.surfout.archive_output_file()
        if self.archive is not None:
            self.archive.archive_files()


# TODO is it used?
def create_working_dir(workdir, enter=True):
    """Create working dir."""
    # Create work directory
    if workdir is not None:
        if os.path.isdir(workdir):
            shutil.rmtree(workdir)
        os.makedirs(workdir, exist_ok=True)
        if enter:
            os.chdir(workdir)


# TODO is it used?
def clean_working_dir(workdir):
    """Clean working dir."""
    if exit:
        os.chdir("..")
    # Clean up
    shutil.rmtree(workdir)



"""Platform dependent settings."""
import json
import logging
import os

from .datetime_utils import as_datetime


class SystemFilePaths(object):
    """Matches files and paths depending on possibly system specific settings.

    User can provide a default system dir to nest dependencies.
    """

    def __init__(self, system_file_paths):
        """Construct SystemFilePaths.

        Args:
            system_file_paths (_type_): _description_
        """
        self.system_file_paths = system_file_paths

    def get_system_path(
        self,
        dtype,
        default_dir=None,
        check_existence=False,
        check_parsing=False,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Get the system path for a given data type.

        Args:
            dtype (str): The data type you want to get the path for (clim_dir/bin_dir etc)
            default_dir (str): A fallback if the desired dtype is not found
            check_existence (bool): Check if the path found also exists
            check_parsing (bool): Check if parsing was successful (all @@ pairs substituted)
            validtime (as_datetime): Parse setting with this as valid time
            basedtg (as_datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member
            tstep (int): Parse setting with this as time step
            pert (int): Parse setting with this as pertubation number
            var (str): Parse setting with this as variable

        Returns:
            data_dir (str):

        Raises:
            RuntimeError: If path not found and check_existence is True

        See Also:
            self.parse_setting

        """
        logging.debug("Search for: %s Default: %s", dtype, str(default_dir))

        data_dir = self.find_matching_data_dir(
            dtype,
            default_dir=default_dir,
            check_existence=check_existence,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )
        if data_dir is None:
            if default_dir is None:
                raise RuntimeError("No system path found for " + dtype)

            logging.debug("Find default path")
            data_dir = self.find_matching_data_dir(
                default_dir,
                default_dir=default_dir,
                check_existence=check_existence,
                check_parsing=check_parsing,
                validtime=validtime,
                basedtg=basedtg,
                mbr=mbr,
                tstep=tstep,
                pert=pert,
                var=var,
            )
            if data_dir is None:
                logging.debug("No default path found for %s", default_dir)

        logging.debug("data_dir %s", data_dir)
        return data_dir

    def find_matching_data_dir(
        self,
        dtype,
        default_dir=None,
        check_existence=False,
        check_parsing=False,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Find a matching path from the system path for a given data type.

        Args:
            dtype (str): The data type you want to get the path for (clim_dir/bin_dir etc)
            default_dir (str): A fallback if the desired dtype is not found
            check_existence (bool): Check if the path found also exists
            check_parsing (bool): Check if parsing was successful (all @@ pairs substituted)
            validtime (as_datetime): Parse setting with this as valid time
            basedtg (as_datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member
            tstep (int): Parse setting with this as time step
            pert (int): Parse setting with this as pertubation number
            var (str): Parse setting with this as variable

        Returns:
            data_dir (str):

        Raises:
            ValueError: data dir is not a string!
            NotADirectoryError: Not a directory

        See Also:
            self.parse_setting

        """
        command = None
        for sfp in self.system_file_paths:
            if sfp == dtype:
                logging.debug("Found %s %s %s", sfp, type(sfp), self.system_file_paths)
                data_dir = self.system_file_paths[sfp]
                # If dict, also a command is attached
                if isinstance(data_dir, dict):
                    for key in data_dir:
                        logging.debug(key, data_dir[key])
                        command = str(data_dir[key])
                        data_dir = str(key)
                logging.debug("Data directory before parsing is: %s", data_dir)
                if not isinstance(data_dir, str):
                    raise ValueError("data dir is not a string!")
                data_dir = self.parse_setting(
                    self.substitute_string(data_dir),
                    check_parsing=check_parsing,
                    validtime=validtime,
                    basedtg=basedtg,
                    mbr=mbr,
                    tstep=tstep,
                    pert=pert,
                    var=var,
                )
                # Add command to data_dir again
                if command is not None:
                    data_dir = {data_dir: command}
                logging.debug("Data directory after parsing is is: %s", data_dir)
                if check_existence:
                    if not os.path.exists(data_dir) and default_dir is None:
                        raise NotADirectoryError(data_dir)
                return data_dir
        return None

    def get_system_file(
        self,
        dtype,
        fname,
        default_dir=None,
        check_existence=False,
        check_parsing=False,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
        system_variables=None,
    ):
        """Get the system path for a given data type and add a file name to the path.

        Args:
            dtype (str): The data type you want to get the path for (clim_dir/bin_dir etc)
            fname (str): Name of the file you want to join to the system path
            default_dir (str): A fallback if the desired dtype is not found
            check_existence (bool): Check if the path found also exists
            check_parsing (bool): Check if parsing was successful (all @@ pairs substituted)
            validtime (as_datetime): Parse setting with this as valid time
            basedtg (as_datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member
            tstep (int): Parse setting with this as time step
            pert (int): Parse setting with this as pertubation number
            var (str): Parse setting with this as variable
            system_variables (dict): Arbitrary settings to substitute @NAME@ =
                                     system_variables={"NAME": "Value"}

        Raises:
            FileNotFoundError: If file not found

        Returns:
            data_dir (str):

        See Also:
            self.parse_setting
            self.substitute_string

        """
        command = None
        path = self.get_system_path(
            dtype,
            default_dir=default_dir,
            check_existence=check_existence,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )

        # If dict, also a command is attached
        if isinstance(path, dict):
            for key in path:
                command = str(path[key])
                path = str(key)
        fname = self.parse_setting(
            fname,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )
        fname = self.substitute_string(fname, system_variables=system_variables)
        if path is None:
            logging.info("No path found for: %s", dtype)
        else:
            fname = path + "/" + fname
        if check_existence:
            if not os.path.exists(fname):
                raise FileNotFoundError(fname)
        if command is not None:
            fname = {fname: command}
        return fname

    @staticmethod
    def parse_setting(
        setting,
        check_parsing=False,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
        micro="@",
    ):
        """Parse setting with date/time/experiment specific values.

        Args:
            setting (str): The value of dictionary key which should be processes. Parser if type is str.
            check_parsing (bool): Check if all @@ pairs were parsed
            validtime (datetime.daetime): Parse setting with this as validtime
            basedtg (datetime.datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member number (@E@/@EE@/@EEE@)
            tstep (int): Parse setting with this as timestep to get step number (@TTT@/@TTTT@)
            pert (int): Parse setting with this as perturbation number @PERT@
            var (str): Parse setting with this as the variable (@VAR@)
            micro (str, optional): Micro character. Defaults to "@"

        Raises:
            RuntimeError: Setting was not substituted properly?

        Returns:
            setting: Possibly parsed setting is type is str

        See Also:
            self.parse_setting
            self.substitute_string

        """
        # Check on arguments
        if isinstance(setting, str):

            if basedtg is not None:
                if isinstance(basedtg, str):
                    basedtg = as_datetime(basedtg)
            if validtime is not None:
                if isinstance(validtime, str):
                    validtime = as_datetime(validtime)
            else:
                validtime = basedtg

            if basedtg is not None and validtime is not None:
                lead_time = validtime - basedtg
                setting = str(setting).replace(
                    f"{micro}YYYY_LL{micro}", validtime.strftime("%Y")
                )
                setting = str(setting).replace(
                    f"{micro}MM_LL{micro}", validtime.strftime("%m")
                )
                setting = str(setting).replace(
                    f"{micro}DD_LL{micro}", validtime.strftime("%d")
                )
                setting = str(setting).replace(
                    f"{micro}HH_LL{micro}", validtime.strftime("%H")
                )
                setting = str(setting).replace(
                    f"{micro}mm_LL{micro}", validtime.strftime("%M")
                )
                setting = str(setting).replace(
                    f"{micro}FG_YYYY{micro}", basedtg.strftime("%Y")
                )
                setting = str(setting).replace(
                    f"{micro}FG_YY{micro}", basedtg.strftime("%y")
                )
                setting = str(setting).replace(
                    f"{micro}FG_MM{micro}", basedtg.strftime("%m")
                )
                setting = str(setting).replace(
                    f"{micro}FG_DD{micro}", basedtg.strftime("%d")
                )
                setting = str(setting).replace(
                    f"{micro}FG_HH{micro}", basedtg.strftime("%H")
                )
                setting = str(setting).replace(
                    f"{micro}FG_mm{micro}", basedtg.strftime("%M")
                )
                lead_seconds = int(lead_time.total_seconds())
                lead_hours = int(lead_seconds / 3600)
                setting = str(setting).replace(f"{micro}LL{micro}", f"{lead_hours:02d}")
                setting = str(setting).replace(f"{micro}LLL{micro}", f"{lead_hours:03d}")
                setting = str(setting).replace(f"{micro}LLLL{micro}", f"{lead_hours:04d}")
                if tstep is not None:
                    lead_step = int(lead_seconds / tstep)
                    setting = str(setting).replace(
                        f"{micro}TTT{micro}", f"{lead_step:03d}"
                    )
                    setting = str(setting).replace(
                        f"{micro}TTTT{micro}", f"{lead_step:04d}"
                    )

            if basedtg is not None:
                setting = str(setting).replace(
                    f"{micro}YMD{micro}", basedtg.strftime("%Y%m%d")
                )
                setting = str(setting).replace(
                    f"{micro}YYYY{micro}", basedtg.strftime("%Y")
                )
                setting = str(setting).replace(
                    f"{micro}YY{micro}", basedtg.strftime("%y")
                )
                setting = str(setting).replace(
                    f"{micro}MM{micro}", basedtg.strftime("%m")
                )
                setting = str(setting).replace(
                    f"{micro}DD{micro}", basedtg.strftime("%d")
                )
                setting = str(setting).replace(
                    f"{micro}HH{micro}", basedtg.strftime("%H")
                )
                setting = str(setting).replace(
                    f"{micro}mm{micro}", basedtg.strftime("%M")
                )

            if mbr is not None:
                setting = str(setting).replace(f"{micro}E{micro}", f"mbr{int(mbr):d}")
                setting = str(setting).replace(f"{micro}EE{micro}", f"mbr{int(mbr):02d}")
                setting = str(setting).replace(f"{micro}EEE{micro}", f"mbr{int(mbr):03d}")
            else:
                setting = str(setting).replace(f"{micro}E{micro}", "")
                setting = str(setting).replace(f"{micro}EE{micro}", "")
                setting = str(setting).replace(f"{micro}EEE{micro}", "")

            if pert is not None:
                logging.debug("replace %s in %s", pert, setting)
                setting = str(setting).replace(f"{micro}PERT{micro}", str(pert))
                logging.debug("replaced %s in %s", pert, setting)

            if var is not None:
                setting = str(setting).replace(f"{micro}VAR{micro}", var)

        if check_parsing:
            if isinstance(setting, str) and setting.count(f"{micro}") > 1:
                raise RuntimeError("Setting was not substituted properly? " + setting)

        return setting

    @staticmethod
    def substitute_string(setting, system_variables=None, micro="@"):
        """Substitute setting if string with OS values of values from system_variables.

        Args:
            setting (str): if setting is string it can be subst
            system_variables (dict): Arbitrary settings to substitute @NAME@ =
                                     system_variables={"NAME": "Value"}
            micro (str, optional): Micro character. Default to "@"

        Returns:
            setting: A setting possibly substituted if type is str

        """
        if isinstance(setting, str):
            env_vals = ["USER", "HOME", "PWD"]
            for env_val in env_vals:
                if env_val in os.environ:
                    setting = setting.replace(
                        f"{micro}" + env_val + f"{micro}", os.environ[env_val]
                    )
                else:
                    logging.debug("%s not found in environment", env_val)

            if system_variables is not None:
                logging.debug(system_variables)
                for var in system_variables:
                    logging.debug(var, system_variables)
                    setting = setting.replace(
                        f"{micro}" + str(var) + f"{micro}", str(system_variables[var])
                    )

        return setting

    def add_system_file_path(
        self,
        name,
        path,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Add a system file path to be used.

        Args:
            name (str): The data type you want to get the path for (clim_dir/bin_dir etc)
            path (str): Name of the file you want to join to the system path
            system_variables (dict): Arbitrary settings to substitute @NAME@ = system_variables=
                                    {"NAME": "Value"}
            check_parsing (bool): Check if parsing was successful (all @@ pairs substituted)
            validtime (as_datetime): Parse setting with this as valid time
            basedtg (as_datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member
            tstep (int): Parse setting with this as time step
            pert (int): Parse setting with this as pertubation number
            var (str): Parse setting with this as variable

        See Also:
            self.parse_setting
            self.substitute_string

        """
        path = self.substitute_string(path, system_variables=system_variables)
        path = self.parse_setting(
            path,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )
        self.system_file_paths.update({name: path})


class SystemFilePathsFromFile(SystemFilePaths):
    """System file paths."""

    def __init__(self, system_file_paths):
        """System file path from a file.

        Args:
            system_file_paths (_type_): _description_

        """
        system_file_paths = json.load(open(system_file_paths, "r", encoding="utf-8"))
        SystemFilePaths.__init__(self, system_file_paths)



"""Configuration."""
import json
import logging
import os

import toml

from .geo import ConfProj
from .platform_deps import SystemFilePaths
from .util import merge_toml_env


class Configuration(object):
    """Configuration class."""

    def __init__(self, conf):
        """Construct the pysurfex configuration.

        Encapsulation of settings dictionary.
        Settings for a deterministic run or an individual member.

        Args:
            conf (dict): Actual psyurfex settings to run

        """
        self.settings = conf
        # Set default file names
        if "CPGDFILE" not in self.settings["SURFEX"]["IO"]:
            self.settings["SURFEX"]["IO"].update({"CPGDFILE": "PGD"})
        if "CPREPFILE" not in self.settings["SURFEX"]["IO"]:
            self.settings["SURFEX"]["IO"].update({"CPREPFILE": "PREP"})
        if "CSURFFILE" not in self.settings["SURFEX"]["IO"]:
            self.settings["SURFEX"]["IO"].update({"CSURFFILE": "SURFOUT"})
        if "LFAGMAP" not in self.settings["SURFEX"]["IO"]:
            self.settings["SURFEX"]["IO"].update({"LFAGMAP": True})

    def dump_json(self, filename, indent=None):
        """Dump configuration to json file.

        Args:
            filename (str): Filename
            indent (int, optional): Indentation. Defaults to None.

        """
        logging.debug("settings %s", self.settings)
        json.dump(self.settings, open(filename, "w", encoding="utf-8"), indent=indent)

    def setting_is(
        self,
        setting,
        value,
        sep="#",
        abort=True,
        default=None,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Check if setting is value.

        Args:
            setting (_type_): _description_
            value (_type_): _description_
            sep (str, optional): _description_. Defaults to "#".
            abort (bool, optional): _description_. Defaults to True.
            default (_type_, optional): _description_. Defaults to None.
            system_variables (_type_, optional): _description_. Defaults to None.
            check_parsing (bool, optional): _description_. Defaults to True.
            validtime (_type_, optional): _description_. Defaults to None.
            basedtg (_type_, optional): _description_. Defaults to None.
            mbr (_type_, optional): _description_. Defaults to None.
            tstep (_type_, optional): _description_. Defaults to None.
            pert (_type_, optional): _description_. Defaults to None.
            var (_type_, optional): _description_. Defaults to None.

        Returns:
            bool: True if found
        """
        if (
            self.get_setting(
                setting,
                sep=sep,
                abort=abort,
                default=default,
                system_variables=system_variables,
                check_parsing=check_parsing,
                validtime=validtime,
                basedtg=basedtg,
                mbr=mbr,
                tstep=tstep,
                pert=pert,
                var=var,
            )
            == value
        ):
            return True
        else:
            return False

    def setting_is_not(
        self,
        setting,
        value,
        sep="#",
        abort=True,
        default=None,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Check if setting is not value.

        Args:
            setting (_type_): _description_
            value (_type_): _description_
            sep (str, optional): _description_. Defaults to "#".
            abort (bool, optional): _description_. Defaults to True.
            default (_type_, optional): _description_. Defaults to None.
            system_variables (_type_, optional): _description_. Defaults to None.
            check_parsing (bool, optional): _description_. Defaults to True.
            validtime (_type_, optional): _description_. Defaults to None.
            basedtg (_type_, optional): _description_. Defaults to None.
            mbr (_type_, optional): _description_. Defaults to None.
            tstep (_type_, optional): _description_. Defaults to None.
            pert (_type_, optional): _description_. Defaults to None.
            var (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        found = False
        if (
            self.get_setting(
                setting,
                sep=sep,
                abort=abort,
                default=default,
                system_variables=system_variables,
                check_parsing=check_parsing,
                validtime=validtime,
                basedtg=basedtg,
                mbr=mbr,
                tstep=tstep,
                pert=pert,
                var=var,
            )
            == value
        ):
            found = True

        if found:
            return False
        else:
            return True

    def value_is_one_of(
        self,
        settings,
        value,
        sep="#",
        abort=True,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Check if value is one of the settings.

        Args:
            settings (list): Settings
            value(any): Value
            sep (str, optional): Separator. Defaults to "#"
            abort (bool, optional): Abort. Defaults to True
            system_variables (dict): Arbitrary settings to substitute
                                     @NAME@ = system_variables={"NAME": "Value"}
            check_parsing (bool): Check if all @@ pairs were parsed
            validtime (as_datetime): Parse setting with this as validtime
            basedtg (as_datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member number (@E@/@EE@/@EEE@)
            tstep (int): Parse setting with this as timestep to get step number (@TTT@/@TTTT@)
            pert (int): Parse setting with this as perturbation number @PERT@
            var (str): Parse setting with this as the variable (@VAR@)

        Returns:
            found (bool): True if value is found in any of the settings

        Raises:
            ValueError: Expected a list as input

        See Also:
            self.get_setting()
            surfex.SystemFilePaths.parse_setting()
            surfex.SystemFilePaths.substitute_string()

        """
        if not isinstance(settings, list):
            raise ValueError("Expected a list as input, got ", type(settings))

        found = False
        for check_s in settings:
            setting = self.get_setting(
                check_s,
                sep=sep,
                abort=abort,
                system_variables=system_variables,
                check_parsing=check_parsing,
                validtime=validtime,
                basedtg=basedtg,
                mbr=mbr,
                tstep=tstep,
                pert=pert,
                var=var,
            )
            if setting == value:
                return True

        return found

    def value_is_not_one_of(
        self,
        setting,
        value,
        sep="#",
        abort=True,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Check if value is not one of.

        Args:
            setting (_type_): _description_
            value (_type_): _description_
            sep (str, optional): _description_. Defaults to "#".
            abort (bool, optional): _description_. Defaults to True.
            system_variables (_type_, optional): _description_. Defaults to None.
            check_parsing (bool, optional): _description_. Defaults to True.
            validtime (_type_, optional): _description_. Defaults to None.
            basedtg (_type_, optional): _description_. Defaults to None.
            mbr (_type_, optional): _description_. Defaults to None.
            tstep (_type_, optional): _description_. Defaults to None.
            pert (_type_, optional): _description_. Defaults to None.
            var (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_

        """
        found = self.value_is_one_of(
            setting,
            value,
            sep=sep,
            abort=abort,
            system_variables=system_variables,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )
        if found:
            return False
        else:
            return True

    def setting_is_one_of(
        self,
        setting,
        values,
        sep="#",
        abort=True,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Check if setting is one of values.

        Args:
            setting (_type_): _description_
            values (_type_): _description_
            sep (str, optional): _description_. Defaults to "#".
            abort (bool, optional): _description_. Defaults to True.
            system_variables (_type_, optional): _description_. Defaults to None.
            check_parsing (bool, optional): _description_. Defaults to True.
            validtime (_type_, optional): _description_. Defaults to None.
            basedtg (_type_, optional): _description_. Defaults to None.
            mbr (_type_, optional): _description_. Defaults to None.
            tstep (_type_, optional): _description_. Defaults to None.
            pert (_type_, optional): _description_. Defaults to None.
            var (_type_, optional): _description_. Defaults to None.

        Raises:
            ValueError: Excpected a list as input

        Returns:
            bool: True if found

        """
        found = False
        setting = self.get_setting(
            setting,
            sep=sep,
            abort=abort,
            system_variables=system_variables,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )
        if not isinstance(values, list):
            raise ValueError("Excpected a list as input, got ", type(values))

        for val in values:
            if setting == val:
                found = True
        return found

    def setting_is_not_one_of(
        self,
        setting,
        values,
        sep="#",
        abort=True,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Check if setting is not one of values.

        Args:
            setting (_type_): _description_
            values (_type_): _description_
            sep (str, optional): _description_. Defaults to "#".
            abort (bool, optional): _description_. Defaults to True.
            system_variables (_type_, optional): _description_. Defaults to None.
            check_parsing (bool, optional): _description_. Defaults to True.
            validtime (_type_, optional): _description_. Defaults to None.
            basedtg (_type_, optional): _description_. Defaults to None.
            mbr (_type_, optional): _description_. Defaults to None.
            tstep (_type_, optional): _description_. Defaults to None.
            pert (_type_, optional): _description_. Defaults to None.
            var (_type_, optional): _description_. Defaults to None.

        Returns:
            bool: _description_

        """
        found = self.setting_is_one_of(
            setting,
            values,
            sep=sep,
            abort=abort,
            system_variables=system_variables,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
        )
        if found:
            return False
        else:
            return True

    def get_setting(
        self,
        setting,
        sep="#",
        abort=True,
        default=None,
        system_variables=None,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
    ):
        """Get configurations setting.

        Settings are nested in blocks.
        To get the full setting request setting joined by a separation character.

        E.g setting = "SURFEX#ASSIM#ASSIM_SCHEMES"

        Args:
            setting (str): The requested setting
            default (any): A fallback setting in case setting is not found
            sep (str): A separation character between different configuration blocks
            abort (bool): Abort if setting is not found and default not set
            system_variables (dict): Arbitrary settings to substitute
                                     @NAME@ = system_variables={"NAME": "Value"}
            check_parsing (bool): Check if all @@ pairs were parsed
            validtime (datetime.daetime): Parse setting with this as validtime
            basedtg (datetime.datetime): Parse setting with this as base time
            mbr (int): Parse setting with this as ensemble member number (@E@/@EE@/@EEE@)
            tstep (int): Parse setting with this as timestep to get step number (@TTT@/@TTTT@)
            pert (int): Parse setting with this as perturbation number @PERT@
            var (str): Parse setting with this as the variable (@VAR@)

        Returns:
            found (bool): True if value is found in any of the settings

        See Also:
            self.get_setting()
            surfex.SystemFilePaths.parse_setting()
            surfex.SystemFilePaths.substitute_string()

        Raises:
            KeyError: Key not found

        """
        settings = self.settings

        if sep is None:
            keys = [setting]
        else:
            keys = setting.split(sep)

        if keys[0] in settings:
            this_setting = settings[keys[0]]
            logging.debug("get_setting %s -> %s", keys[0], this_setting)
            if len(keys) > 1:
                for key in keys[1:]:
                    if key in this_setting:
                        this_setting = this_setting[key]
                        # Time information
                        this_setting = SystemFilePaths.substitute_string(
                            this_setting, system_variables=system_variables
                        )
                        this_setting = SystemFilePaths.parse_setting(
                            this_setting,
                            check_parsing=check_parsing,
                            validtime=validtime,
                            basedtg=basedtg,
                            mbr=mbr,
                            tstep=tstep,
                            pert=pert,
                            var=var,
                        )
                    else:
                        if default is not None:
                            this_setting = default
                        elif abort:
                            raise KeyError("Key not found " + key)
                        else:
                            this_setting = None
                            break
        else:
            if abort:
                raise KeyError("Key not found " + keys[0])
            else:
                this_setting = None

        logging.debug("get_setting %s %s %s", setting, this_setting, type(this_setting))
        return this_setting

    def update_setting(self, setting, value, sep="#"):
        """Update the setting.

        Args:
            setting (_type_): _description_
            value (_type_): _description_
            sep (str, optional): _description_. Defaults to "#".
        """
        if sep is None:
            keys = [setting]
        else:
            keys = setting.split(sep)

        last_key = keys[-1]
        dsetting = {last_key: value}
        if len(keys) > 1:
            for key in reversed(keys[0:-1]):
                dsetting = {key: dsetting}

        self.settings = merge_toml_env(self.settings, dsetting)


class ConfigurationFromHarmonie(Configuration):
    """Set the configuration from Harmonie environment.

    This class sets up a SURFEX configuration from an environment based Harmonie run with it's
    corresponding configuration.

    Some settings imply several changes in SURFEX configuration

    """

    def __init__(self, env, conf):
        """Constuct the Configuration object.

        Args:
            env (dict): System environment e.g. os.environ
            conf (dict): The default configuration for this deterministic run/ensemble member

        Raises:
            NotImplementedError: soil_texture not implemented

        """
        Configuration.__init__(self, conf)

        # Set domain from environment variables. Geo is alway conf proj
        ezone = int(env["EZONE"])
        ndluxg = int(env["NLON"]) - ezone
        if "LNDLUXG" in env:
            ndluxg = int(env["LNDLUXG"])
        ndguxg = int(env["NLAT"]) - ezone
        if "LNDGUXG" in env:
            ndguxg = int(env["LNDGUXG"])
        gsize = float(env["GSIZE"])
        if "LGSIZE" in env:
            gsize = float(env["LGSIZE"])
        trunc = 2  # linear
        if "TRUNC" in env:
            trunc = int(env["TRUNC"])
        domain_dict = {
            "nam_pgd_grid": {"cgrid": "CONF PROJ"},
            "nam_conf_proj": {
                "xlat0": float(env["LAT0"]),
                "xlon0": float(env["LON0"]),
            },
            "nam_conf_proj_grid": {
                "ilone": ezone,
                "ilate": ezone,
                "xlatcen": float(env["LATC"]),
                "xloncen": float(env["LONC"]),
                "nimax": ndluxg,
                "njmax": ndguxg,
                "xdx": gsize,
                "xdy": gsize,
                "xtrunc": trunc,
            },
        }
        geo = ConfProj(domain_dict)
        self.geo = geo

        logging.debug("GEO: %s", self.geo)

        self.settings.update({"FORECAST": {"PHYSICS": env["PHYSICS"]}})

        # Ensemble member information
        mbr = None
        if "ENSMBR" in env:
            mbr = env["ENSMBR"]
        self.settings.update({"EPS": {"ENSMBR": mbr}})

        # IO
        cnmexp = os.environ["CNMEXP"]
        self.update_setting("SURFEX#IO#CPGDFILE", "Const.Clim")
        self.update_setting("SURFEX#IO#CPREPFILE", "ICMSH" + cnmexp + "INIT")
        self.update_setting("SURFEX#IO#CSURFFILE", "ICMSH" + cnmexp + "+@LLLL@")
        self.update_setting("SURFEX#IO#CSURF_FILETYPE", "FA")
        self.update_setting("SURFEX#IO#CTIMESERIES_FILETYPE", "FA")
        self.update_setting("SURFEX#IO#LFAGMAP", True)
        lselect = False
        if os.environ["SURFEX_LSELECT"] == "yes":
            lselect = True
        self.update_setting("SURFEX#IO#LSELECT", lselect)

        #  CISBA Type of ISBA scheme in SURFEX. Options: "3-L"|"2-L"|"DIF"
        self.update_setting("SURFEX#ISBA#SCHEME", env["CISBA"])
        if env["CISBA"] == "DIF":
            self.update_setting("SURFEX#ISBA#MEB", True)

        # CSNOW Type of snow scheme in SURFEX. Options: "D95" and "3-L"
        self.update_setting("SURFEX#ISBA#SNOW", env["CSNOW"])

        # NPATCH Number of patches over land in SURFEX (see also LISBA_CANOPY)
        self.update_setting("SURFEX#ISBA#NPATCH", int(env["NPATCH"]))

        # LISBA_CANOPY Activates surface boundary multi layer scheme over land in SURFEX
        # (must be .FALSE. for NPATCH>1)
        canopy = env["LISBA_CANOPY"].replace(".", "")
        if canopy.strip().lower()[0] == "t":
            canopy = True
        else:
            canopy = False
        self.update_setting("SURFEX#ISBA#CANOPY", canopy)

        # CROUGH SSO scheme used in SURFEX "NONE"|"Z01D"|"BE04"|"OROT"
        self.update_setting("SURFEX#SSO#SCHEME", env["CROUGH"])

        # SURFEX_SEA_ICE Treatment of sea ice in surfex (none|sice)
        self.update_setting("SURFEX#SEA#ICE", env["SURFEX_SEA_ICE"].upper())

        # SURFEX_LAKES Treatment of lakes in surfex (WATFLX|FLAKE)
        self.update_setting("SURFEX#TILES#INLAND_WATER", env["SURFEX_LAKES"])

        # TOPO_SOURCE Input source for orography. Available are (gmted2010|gtopo30)
        self.update_setting("SURFEX#ZS#YZS", env["TOPO_SOURCE"] + ".dir")

        # ECOCLIMAP
        ecoclimap_version = env["ECOCLIMAP_VERSION"]
        if ecoclimap_version == "SG":
            self.update_setting("SURFEX#COVER#SG", True)
            self.update_setting("SURFEX#COVER#YCOVER", "ecosg_final_map.dir")
        else:
            self.update_setting("SURFEX#COVER#SG", False)
            version1 = ["1.0", "1.2", "1.3", "1.4", "1.5"]
            version2 = ["2.0", "2.1", "2.2", "2.2.1", "2.5_plus"]
            if ecoclimap_version in version1:
                ycover = "ECOCLIMAP_I_GLOBAL"
            elif ecoclimap_version in version2:
                ycover = "ECOCLIMAP_II_EUROP"
            else:
                raise NotImplementedError
            ycover_ver = ycover + "_V" + ecoclimap_version
            self.update_setting("SURFEX#COVER#YCOVER", ycover_ver + ".dir")

        # SOIL_TEXTURE_VERSION  # Soil texture input data FAO|HWSD_v2|SOILGRID
        # Soil texture (sand/clay)
        soil_texture = env["SOIL_TEXTURE_VERSION"]
        if soil_texture == "FAO":
            ysand = "sand_fao"
            yclay = "clay_fao"
        elif soil_texture == "HWSD_v2":
            ysand = "SAND_HWSD_MOY_v2"
            yclay = "CLAY_HWSD_MOY_v2"
        elif soil_texture == "SOILGRID":
            ysand = "SAND_SOILGRID"
            yclay = "CLAY_SOILGRID"
        elif soil_texture == "SOILGRID_v2":
            ysand = "sand_0-200cm_mean_int"
            yclay = "clay_0-200cm_mean_int"
        else:
            raise NotImplementedError
        self.update_setting("SURFEX#ISBA#YSAND", ysand + ".dir")
        self.update_setting("SURFEX#ISBA#YCLAY", yclay + ".dir")

        # Lake database version.
        self.update_setting("SURFEX#FLAKE#LDB_VERSION", env["LDB_VERSION"])

        # Treeheight
        if "H_TREE_FILE" in env:
            self.update_setting("SURFEX#COVER#H_TREE", env["H_TREE_FILE"])

        # XRIMAX Maximum allowed Richardson number in the surface layer (cy40h default was 0.0)
        self.update_setting("SURFEX#PARAMETERS#XRIMAX", float(env["XRIMAX"]))

        # XSCALE_H_TREE  Scale the tree height with this factor
        self.update_setting("SURFEX#TREEDRAG#XSCALE_H_TREE", env["XSCALE_H_TREE"])
        if "LFAKETREE" in env:
            if env["LFAKETREE"].replace(".", "").strip().lower()[0] == "t":
                lfaketree = True
            else:
                lfaketree = False
            self.update_setting("SURFEX#TREEDRAG#FAKETREES", lfaketree)

        # Heat capacity
        if "XCGMAX" in env:
            self.update_setting("SURFEX#ISBA#XCGMAX", float(env["XCGMAX"]))
        if "XCSMAX" in env:
            self.update_setting("SURFEX#ISBA#XCSMAX", float(env["XCSMAX"]))

        # CFORCING_FILETYPE Offline surfex forcing format (NETCDF/ASCII)
        self.update_setting("SURFEX#IO#CFORCING_FILETYPE", env["CFORCING_FILETYPE"])

        #########################
        # Assimilation settings #
        #########################
        # Default schemes. Possible to override in future
        # Sea
        ana_sea = "INPUT"
        if "ANA_SEA" in env:
            ana_sea = env["ANA_SEA"]
        self.update_setting("SURFEX#ASSIM#SCHEMES#SEA", ana_sea)

        if "LECSST" in env:
            if env["LECSST"].replace(".", "").strip().lower()[0] == "t":
                lecsst = True
            else:
                lecsst = False
            self.update_setting("SURFEX#ASSIM#SEA#LECSST", lecsst)
        else:
            self.update_setting("SURFEX#ASSIM#SEA#LECSST", True)

        # Inland water
        ana_lake = "INPUT"
        if "ANA_LAKE" in env:
            ana_lake = env["ANA_LAKE"]
        self.update_setting("SURFEX#ASSIM#SCHEMES#INLAND_WATER", ana_lake)
        # TEB
        ana_teb = "ROADT"
        if "ANA_TEB" in env:
            ana_teb = env["ANA_TEB"]
        self.update_setting("SURFEX#ASSIM#SCHEMES#TEB", ana_teb)

        # Soil assimilation
        anasurf = env["ANASURF"]
        if anasurf == "OI" or anasurf == "CANARI_OI_MAIN":
            self.update_setting("SURFEX#ASSIM#SCHEMES#ISBA", "OI")
        if anasurf == "EKF" or anasurf == "CANARI_EKF_SURFEX":
            self.update_setting("SURFEX#ASSIM#SCHEMES#ISBA", "EKF")
        if anasurf == "ENKF":
            self.update_setting("SURFEX#ASSIM#SCHEMES#ISBA", "ENKF")

        # Active EKF control variables from CVAR_M
        if "NNCV" in env:
            nncv = env["NNCV"]
            nncv = list(map(int, nncv.split(",")))
            self.update_setting("SURFEX#ASSIM#ISBA#EKF#NNCV", nncv)

        if "CVAR_M" in env:
            cvar_m = env["CVAR_M"]
            cvar_m = list(map(str, cvar_m.split(",")))
            self.update_setting("SURFEX#ASSIM#ISBA#EKF#CVAR_M", cvar_m)

        if "XSIGMA_M" in env:
            xsigma_m = env["XSIGMA_M"]
            xsigma_m = list(map(float, xsigma_m.split(",")))
            self.update_setting("SURFEX#ASSIM#ISBA#EKF#XSIGMA_M", xsigma_m)

        if "XTPRT_M" in env:
            xtprt_m = env["XTPRT_M"]
            xtprt_m = list(map(float, xtprt_m.split(",")))
            self.update_setting("SURFEX#ASSIM#ISBA#EKF#XTPRT_M", xtprt_m)

        if "LLINCHECK" in env:
            llincheck = env["LLINCHECK"]
            if llincheck == "TRUE":
                self.update_setting("SURFEX#ASSIM#ISBA#EKF#LLINCHECK", True)
            else:
                self.update_setting("SURFEX#ASSIM#ISBA#EKF#LLINCHECK", False)

        if "XALPHA" in env:
            xalpha = env["XALPHA"]
            self.update_setting("SURFEX#ASSIM#ISBA#EKF#XALPHA", float(xalpha))

        # Observations
        nnco = env["NNCO"]
        nnco = list(map(int, nnco.split(",")))
        self.update_setting("SURFEX#ASSIM#OBS#NNCO", nnco)
        if "COBS_M" in env:
            cobs_m = env["COBS_M"]
            cobs_m = list(map(str, cobs_m.split(",")))
            self.update_setting("SURFEX#ASSIM#OBS#COBS_M", cobs_m)
        if "XERROBS_M" in env:
            xerrobs_m = env["XERROBS_M"]
            xerrobs_m = list(map(float, xerrobs_m.split(",")))
            self.update_setting("SURFEX#ASSIM#OBS#XERROBS_M", xerrobs_m)

        # Active EnKF control variables from CVAR_M
        if "NNCV" in env:
            nncv = env["NNCV"]
            nncv = list(map(int, nncv.split(",")))
            self.update_setting("SURFEX#ASSIM#ISBA#ENKF#NNCV", nncv)

        if "CVAR_M" in env:
            cvar_m = env["CVAR_M"]
            cvar_m = list(map(str, cvar_m.split(",")))
            self.update_setting("SURFEX#ASSIM#ISBA#ENKF#CVAR_M", cvar_m)

        if "NENS_M" in env:
            nens_m = env["NENS_M"]
            self.update_setting("SURFEX#ASSIM#ISBA#ENKF#NENS_M", int(nens_m))

        # ANASURF_OI_COEFF Specify use of OI coefficients file (POLYNOMES_ISBA|POLYNOMES_ISBA_MF6)
        # # POLYNOMES_ISBA_MF6 means 6 times smaller coefficients for WG2 increments
        self.update_setting("SURFEX#ASSIM#ISBA#OI#COEFFS", env["ANASURF_OI_COEFF"])

        # Always set SURFEX#IO#LSPLIT_PATCH False
        self.update_setting("SURFEX#IO#LSPLIT_PATCH", False)

        # Always use FA format as input
        self.update_setting("SURFEX#ASSIM#CFILE_FORMAT_LSM", "FA")
        self.update_setting("SURFEX#ASSIM#SEA#CFILE_FORMAT_SST", "FA")
        self.update_setting("SURFEX#ASSIM#ISBA#OI#CFILE_FORMAT_FG", "FA")
        self.update_setting("SURFEX#ASSIM#ISBA#OI#CFILE_FORMAT_CLIM", "FA")
        if anasurf == "CANARI_OI_MAIN" or anasurf == "CANARI_EKF_SURFEX":
            self.update_setting("SURFEX#ASSIM#OBS#CFILE_FORMAT_OBS", "FA")
            self.update_setting("SURFEX#ASSIM#OBS#LSWE", True)
        else:
            self.update_setting("SURFEX#ASSIM#OBS#CFILE_FORMAT_OBS", "ASCII")
            self.update_setting("SURFEX#ASSIM#OBS#LSWE", False)

        snow_cycles = ["06"]
        if "SNOW_CYCLES" in env:
            if (env["SNOW_CYCLES"]) == "":
                snow_cycles = []
            else:
                snow_cycles = str(env["SNOW_CYCLES"]).split(" ")
        self.update_setting("SURFEX#ASSIM#ISBA#UPDATE_SNOW_CYCLES", snow_cycles)

        lswepsini = False
        if "LSWEPSINI" in env:
            if env["LSWEPSINI"].replace(".", "").strip().lower()[0] == "t":
                lswepsini = True
            else:
                lswepsini = False
        self.update_setting("SURFEX#ASSIM#ISBA#LSWEPSINI", lswepsini)
        xswepsini = 1000.0
        if "XSWEPSINI" in env:
            xswepsini = float(env["XSWEPSINI"])
        self.update_setting("SURFEX#ASSIM#ISBA#XSWEPSINI", xswepsini)
        lswepsmin = False
        if "LSWEPSMIN" in env:
            if env["LSWEPSMIN"].replace(".", "").strip().lower()[0] == "t":
                lswepsmin = True
            else:
                lswepsmin = False
        self.update_setting("SURFEX#ASSIM#ISBA#LSWEPSMIN", lswepsmin)
        xswepsmin = 500.0
        if "XSWEPSMIN" in env:
            xswepsmin = float(env["XSWEPSMIN"])
        self.update_setting("SURFEX#ASSIM#ISBA#XSWEPSMIN", xswepsmin)

        lpatch1 = False
        if "LPATCH1" in env:
            if env["LPATCH1"].replace(".", "").strip().lower()[0] == "t":
                lpatch1 = True
            else:
                lpatch1 = False
        self.update_setting("SURFEX#ASSIM#ISBA#LPATCH1", lpatch1)

        # Perturbations
        # PERTSURF ECMA    : perturb also the surface observation before Canari (recommended
        #                  : for EDA to have full perturbation of the initial state).
        #          model   : perturb surface fields in grid-point space (recursive filter)
        #          none    : no perturbation for surface observations.
        self.update_setting("SURFEX#ISBA#PERTSURF", False)
        self.update_setting("SURFEX#SEA#PERTFLUX", False)
        if env["PERTSURF"] == "model":
            if "LPERTSURF" in env:
                if env["LPERTSURF"].replace(".", "").strip().lower()[0] == "t":
                    self.update_setting("SURFEX#ISBA#PERTSURF", True)
                    self.update_setting("SURFEX#SEA#PERTFLUX", True)

        # Volatile sea ice (climate mode)
        if "LVOLATILE_SIC" in env:
            if env["LVOLATILE_SIC"].replace(".", "").strip().lower()[0] == "t":
                self.update_setting("SURFEX.SEA.LVOLATILE_SIC", True)
            else:
                self.update_setting("SURFEX.SEA.LVOLATILE_SIC", False)


class ConfigurationFromHarmonieAndConfigFile(ConfigurationFromHarmonie):
    """Initialize a configuration from envrionment and a toml configuration file."""

    def __init__(self, env, conf_file):
        """Initialize a configuration from envrionment and a toml configuration file.

        Args:
            env (dict): System environment e.g. os.environ
            conf_file (str): Filename with configuration

        """
        with open(conf_file, "r", encoding="utf-8") as fhandler:
            conf = toml.load(fhandler)
        ConfigurationFromHarmonie.__init__(self, env, conf)


class ConfigurationFromTomlFile(Configuration):
    """Configuration from a TOML file."""

    def __init__(self, filename):
        """Construct the configuration.

        Args:
            filename (str): File name

        """
        with open(filename, mode="r", encoding="utf-8") as fhandler:
            settings = toml.load(fhandler)
        Configuration.__init__(self, settings)



"""Interpolation. All interfaces to gridpp."""
import logging

try:
    import gridpp
except ModuleNotFoundError:
    gridpp = None
import numpy as np


class Grid:
    """A gridpp Grid wrapper.

    Gridpp expects lat, lon, but pysurfex uses lon, lat.
    This class handles the conversion for the grid

    """

    def __init__(self, grid_lons, grid_lats, elevs=None):
        """Construct a gridpp Grid wrapper.

        Args:
            grid_lons (np.ndarray): grid longitudes
            grid_lats (np.ndarray): grid latitudes
            elevs (np.ndarray, optional): elevations

        Raises:
            RuntimeError: You need gridpp for interpolation

        """
        if gridpp is None:
            raise RuntimeError("You need gridpp for interpolation")
        logging.debug("gridpp.__file__ = %s", gridpp.__file__)
        logging.debug("grid_lons_shape in 0 %s", grid_lons.shape)
        logging.debug("grid_lats_shape in 0 %s", grid_lats.shape)
        grid_lons = np.transpose(grid_lons)
        grid_lats = np.transpose(grid_lats)
        logging.debug("grid_lons_shape in 1 %s", grid_lons.shape)
        logging.debug("grid_lats_shape in 1 %s", grid_lats.shape)
        if elevs is None:
            self.grid = gridpp.Grid(grid_lats, grid_lons)
        else:
            elevs = np.transpose(elevs)
            self.grid = gridpp.Grid(grid_lats, grid_lons, elevs)


class Points:
    """A gridpp Points wrapper.

    Gridpp expects lat, lon, but pysurfex uses lon, lat.
    This class handles the conversion for the grid

    """

    def __init__(self, p_lons, p_lats, p_elevs=None):
        """Construct a gridpp Points wrapper.

        Args:
            p_lons(np.array): Point longitudes
            p_lats(np.array): Point latitudes
            p_elevs(np.array, optional): Point elevations. Defaults to None.

        Raises:
            RuntimeError: You need gridpp for interpolation

        """
        if gridpp is None:
            raise RuntimeError("You need gridpp for interpolation")
        if isinstance(p_lons, list):
            p_lons = np.asarray(p_lons)
        if isinstance(p_lats, list):
            p_lats = np.asarray(p_lats)
        logging.debug("p_lons_shape in %s", p_lons.shape)
        logging.debug("p_lats_shape in %s", p_lats.shape)
        self.lons = np.transpose(p_lons)
        self.lats = np.transpose(p_lats)
        self.elevs = p_elevs
        logging.debug("self.lons_shape in %s", self.lons.shape)
        logging.debug("self.lats_shape in %s", self.lats.shape)
        if p_elevs is None:
            self.points = gridpp.Points(self.lats, self.lons)
        else:
            p_elevs = np.transpose(p_elevs)
            self.points = gridpp.Points(self.lats, self.lons, p_elevs)

    def inside_grid(self, grid, distance=2500.0):
        """Check if inside grid.

        Args:
            grid(Grid): Grid object
            distance(float, optional): Max distance from grid. Defaults to 2500.0.

        Returns:
            inside_grid(list): Logical mask if inside grid.
        """
        inside_grid = []
        # Check if they are in grid
        for i in range(0, self.lons.shape[0]):
            lon = self.lons[i]
            lat = self.lats[i]
            neighbours = grid.grid.get_num_neighbours(lat, lon, distance)
            if neighbours == 0:
                inside_grid.append(False)
            else:
                inside_grid.append(True)
        return inside_grid


class Interpolation(object):
    """Interpolation."""

    def __init__(self, operator, geo_in, geo_out):
        """Construct an intrpolation object.

        Args:
            operator (str): Operator
            geo_in (surfex.geo.Geo): Input geometry
            geo_out (surfex.geo.Geo): Output geometry

        Raises:
            RuntimeError: You can not interpolate without specifying an output geometry

        """
        self.operator = operator
        self.geo_in = geo_in
        self.geo_out = geo_out
        if self.geo_out is None:
            raise RuntimeError(
                "You can not interpolate without specifying an output geometry"
            )
        # Input
        if self.geo_in is not None:
            logging.debug("grid_lons.shape in %s", geo_in.lons.shape)
            logging.debug("grid_lats.shape in %s", geo_in.lats.shape)
            self.var_lons = geo_in.lons
            self.var_lats = geo_in.lats
            self.identical = self.geo_out.is_identical(self.geo_in)
        else:
            self.var_lons = None
            self.var_lats = None
            self.identical = False

        # Output
        self.lons = self.geo_out.lonlist
        self.lats = self.geo_out.latlist
        self.npoints = self.geo_out.npoints

    def interpolate(self, field2d, undefined=None):
        """Do interpolation.

        Args:
            field2d (np.ndarray): Two dimensional field to interpolate
            undefined (float, optional): Undefined value if field2d is None. Defaults to None.

        Raises:
            RuntimeError: Input domain and ouput domain differ.
            RuntimeError: You try to interpolate a missing field!
            NotImplementedError: Interpolation method not implemented

        Returns:
            np.array: interpolated_field

        """
        if field2d is None and undefined is not None:
            return np.full((self.geo_out.nlons * self.geo_out.nlats), undefined)
        elif field2d is None:
            raise RuntimeError("You try to interpolate a missing field!")
        else:
            logging.debug("field2d.shape %s", field2d.shape)
            if self.identical or self.operator == "none":
                if self.operator == "none":
                    if not self.identical:
                        raise RuntimeError(
                            "Input domain and ouput domain differ. "
                            "You must interpolate!"
                        )
                    logging.info("No interpolation chosen")
                else:
                    logging.info(
                        "Input and output domain are identical. "
                        "No interpolation is needed"
                    )
                interpolated_field = field2d.reshape(self.npoints)
            else:
                sub_lons, sub_lats = self.geo_out.subset(self.geo_in)
                if len(sub_lons) == 0 and len(sub_lats) == 0:
                    logging.info(
                        "Doing '%s' interpolation for %s points",
                        self.operator,
                        str(self.npoints),
                    )
                    if self.operator == "nearest" or self.operator == "bilinear":
                        grid = Grid(self.var_lons, self.var_lats)
                        points = Points(self.lons, self.lats)
                        interpolated_field = grid2points(
                            grid, points, field2d, operator=self.operator
                        )
                    elif self.operator == "none":
                        interpolated_field = field2d.reshape(self.npoints)
                    else:
                        raise NotImplementedError(self.operator)
                else:
                    logging.info("Output domain is a subset of input domain")
                    new_field = np.ndarray([len(sub_lons), len(sub_lats)])
                    for i, sub_lon in enumerate(sub_lons):
                        for j, sub_lat in enumerate(sub_lats):
                            new_field[i, j] = field2d[sub_lon, sub_lat]
                    interpolated_field = new_field.reshape(self.npoints)
            return interpolated_field

    def rotate_wind_to_geographic(self):
        """Not implemented."""

    @staticmethod
    def distance(lon1, lat1, lon2, lat2):
        """Compute distance.

        Computes the great circle distance between two points using the
        haversine formula. Values can be vectors.

        Args:
            lon1(float): lon1
            lat1(float): lat1
            lon2(float): lon2
            lat2(float): lat2

        Returns:
            cval(float): Circle distance

        """
        # Convert from degrees to radians
        pi_constant = 3.14159265
        lon1 = lon1 * 2 * pi_constant / 360.0
        lat1 = lat1 * 2 * pi_constant / 360.0
        lon2 = lon2 * 2 * pi_constant / 360.0
        lat2 = lat2 * 2 * pi_constant / 360.0
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        aval = (
            np.sin(dlat / 2.0) ** 2
            + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2
        )
        cval = 2 * np.arcsin(np.sqrt(aval)) * 6.367e6
        return cval

    def alpha_grid_rot(self):
        """Calculate alpha."""
        lon = self.var_lons
        lat = self.var_lats
        n_x = lat.shape[0]
        dlon = np.zeros(lat.shape)
        dlat = np.zeros(lat.shape)
        i_1 = np.arange(n_x - 1)

        dlon[0:-1, :] = np.sign(lon[i_1 + 1, :] - lon[i_1, :]) * self.distance(
            lon[i_1, :], lat[i_1, :], lon[i_1 + 1, :], lat[i_1, :]
        )
        dlat[0:-1, :] = -np.sign(lat[i_1 + 1, :] - lat[i_1, :]) * self.distance(
            lon[i_1, :], lat[i_1, :], lon[i_1, :], lat[i_1 + 1, :]
        )

        dlon[-1, :] = np.sign(lon[-1, :] - lon[-2, :]) * self.distance(
            lon[-2, :], lat[-2, :], lon[-1, :], lat[-2, :]
        )
        dlat[-1, :] = -np.sign(lat[-1, :] - lat[-2, :]) * self.distance(
            lon[-2, :], lat[-2, :], lon[-2, :], lat[-1, :]
        )

        alpha = np.rad2deg(np.arctan2(dlon, dlat))
        return alpha


def fill_field(field_tmp, geo, radius=1):
    """Fill field.

    Args:
        field_tmp (np.ndarray): Field
        geo (surfex.geo.Geo): Geometry
        radius (int, optional): Radius. Defaults to 1.

    Returns:
        tuple: field, nans

    Raises:
        RuntimeError: You need gridpp for fill_field

    """
    if gridpp is None:
        raise RuntimeError("You need gridpp for fill_field")
    ovalues = gridpp.neighbourhood(field_tmp, radius, gridpp.Mean)
    nans = 0
    for i in range(0, geo.nlons):
        for j in range(0, geo.nlats):
            if np.isnan(field_tmp[i][j]):
                nans = nans + 1
                field_tmp[i][j] = ovalues[i][j]
    return field_tmp, nans


def gridpos2points(
    grid_lons,
    grid_lats,
    p_lons,
    p_lats,
    grid_values,
    operator="bilinear",
    elev_gradient=None,
):
    """Convert grid positions to points.

    Args:
        grid_lons (np.ndarray): Grid longitudes
        grid_lats (np.ndarray): Grid latitudes
        p_lons (np.ndaray): Point longitudes
        p_lats (np.ndarray): Point latitudes
        grid_values (np.ndarray): Grid values
        operator (str, optional): Interpolation operator. Defaults to "bilinear".
        elev_gradient (float, optional): Elevation gradient for downscaler

    Returns:
        np.ndarray: Interpolated values
    """
    grid = Grid(grid_lons, grid_lats)
    points = Points(p_lons, p_lats)
    return grid2points(
        grid, points, grid_values, operator=operator, elev_gradient=elev_gradient
    )


def grid2points(grid, points, grid_values, operator="bilinear", elev_gradient=None):
    """Convert a grid to points.

    Args:
        grid (Grid): Grid object
        points (Points): Points object
        grid_values (np.ndarray): Grid values
        operator (str, optional): Interpolation operator. Defaults to "bilinear".
        elev_gradient (float, optional): Elevation gradient for downscaler

    Raises:
        NotImplementedError: Operator not implemented
        RuntimeError: You need gridpp for interpolation

    Returns:
        np.ndarray: Interpolated values
    """
    if gridpp is None:
        raise RuntimeError("You need gridpp for interpolation")
    logging.debug("grid_values.shape 0: %s", grid_values.shape)
    grid_values = np.transpose(grid_values)
    logging.debug("grid_values.shape 1: %s", grid_values.shape)
    if operator == "bilinear":
        if elev_gradient is None:
            values = gridpp.bilinear(grid.grid, points.points, grid_values)
        else:
            values = gridpp.simple_gradient(
                grid.grid, points.points, grid_values, elev_gradient, gridpp.Bilinear
            )
    elif operator == "nearest":
        if elev_gradient is None:
            values = gridpp.nearest(grid.grid, points.points, grid_values)
        else:
            values = gridpp.simple_gradient(
                grid.grid, points.points, grid_values, elev_gradient, gridpp.Nearest
            )
    else:
        raise NotImplementedError(f"Operator {operator} not implemented!")
    return values


def inside_grid(grid_lons, grid_lats, p_lons, p_lats, distance=2500.0):
    """Get number of neighbours.

    Args:
        grid_lons (np.ndarray): Grid longitudes
        grid_lats (np.ndarray): Grid latitudes
        p_lons (np.array): Point longitudes
        p_lats (np.array): Point latitudes
        distance (float, optional): Max distance from points. Defaults to 2500.0.

    Returns:
        inside_grid(list): Boolean mask

    """
    grid = Grid(grid_lons, grid_lats)
    points = Points(p_lons, p_lats)
    return points.inside_grid(grid, distance=distance)


class ObsOperator(object):
    """Obs operator. Class to convert a field to an observation point."""

    def __init__(self, operator, geo, dataset, grid_values, max_distance=5000):
        """Construct the observation operator.

        Args:
            operator (str): Interpolation operator.
            geo (surfex.Geo): Surfex geometry.
            dataset (QCDataSet): QC data set.
            grid_values (np.darray): Values in the grid.
            max_distance (int, optional): Max allowed deviation in meters from grid borders.
                                          Defaults to 5000.

        """
        lons = dataset.lons
        lats = dataset.lats
        logging.info(
            'Setting up "%s" observation operator for %s points', operator, str(len(lons))
        )
        obs_values = gridpos2points(
            geo.lons, geo.lats, lons, lats, grid_values, operator=operator
        )
        self.inside_grid = inside_grid(
            geo.lons, geo.lats, lons, lats, distance=max_distance
        )
        self.obs_values = obs_values

    def get_obs_value(self, pos=None):
        """Get the observed value.

        Args:
            pos (int, optional): Position. Defaults to None.

        Raises:
            NotImplementedError: Specific position not implemented yet!

        Returns:
            float: Observation value for index.
        """
        if pos is None:
            return self.obs_values
        else:
            raise NotImplementedError("Specific position not implemented yet!")

    def is_in_grid(self, index):
        """Check if index is in grid.

        Args:
            index (int): Index to check.

        Returns:
            bool: True if inside
        """
        return self.inside_grid[index]


def horizontal_oi(
    geo,
    background,
    observations,
    gelevs,
    hlength=10000.0,
    vlength=10000.0,
    wlength=0.5,
    elev_gradient=None,
    structure_function="Barnes",
    max_locations=50,
    epsilon=0.5,
    minvalue=None,
    maxvalue=None,
    interpol="bilinear",
    only_diff=False,
):
    """Do horizontal OI.

    Args:
        geo (_type_): _description_
        background (_type_): _description_
        observations (_type_): _description_
        gelevs (_type_): _description_
        hlength (_type_, optional): _description_. Defaults to 10000..
        vlength (_type_, optional): _description_. Defaults to 10000..
        wlength (float, optional): _description_. Defaults to 0.5.
        elev_gradient (int, optional): _description_. Defaults to 0.
        structure_function (str, optional): _description_. Defaults to "Barnes".
        max_locations (int, optional): _description_. Defaults to 50.
        epsilon (float, optional): _description_. Defaults to 0.5.
        minvalue (_type_, optional): _description_. Defaults to None.
        maxvalue (_type_, optional): _description_. Defaults to None.
        interpol (str, optional): _description_. Defaults to "bilinear".
        only_diff (bool, optional): _description_. Defaults to False.

    Raises:
        NotImplementedError: Structure function not implemented
        RuntimeError: You need gridpp to perform OI

    Returns:
        _type_: _description_

    """
    if gridpp is None:
        raise RuntimeError("You need gridpp to perform OI")

    logging.debug(gridpp.__file__)
    logging.debug(gridpp.__version__)
    glats = geo.lats
    glons = geo.lons

    def obs2vectors(my_obs):
        return (
            my_obs.lons,
            my_obs.lats,
            my_obs.stids,
            my_obs.elevs,
            my_obs.values,
            my_obs.cis,
            my_obs.lafs,
        )

    vectors = np.vectorize(obs2vectors)
    lons, lats, __, elevs, values, __, __ = vectors(observations)

    bgrid = Grid(glons, glats, gelevs)
    points = Points(lons, lats, elevs)
    pbackground = grid2points(
        bgrid, points, background, operator=interpol, elev_gradient=elev_gradient
    )

    # Remove undefined backgrounds
    if any(np.isnan(pbackground)):
        print("Found undefined backgrounds. Remove them")
        lons2 = []
        lats2 = []
        elevs2 = []
        values2 = []
        for point in range(0, len(lons)):
            if np.isnan(pbackground[point]):
                logging.info(
                    "Undefined background in lon=%s lat=%s value=%s",
                    lons[point],
                    lats[point],
                    values[point],
                )
            else:
                lons2.append(lons[point])
                lats2.append(lats[point])
                elevs2.append(elevs[point])
                values2.append(values[point])
        values = values2
        points = Points(lons2, lats2, elevs2)
        pbackground = grid2points(
            bgrid, points, background, operator=interpol, elev_gradient=elev_gradient
        )

    variance_ratios = np.full(points.points.size(), epsilon)

    if structure_function == "Barnes":
        structure = gridpp.BarnesStructure(hlength, vlength, wlength)
    else:
        raise NotImplementedError

    background = np.transpose(background)
    field = gridpp.optimal_interpolation(
        bgrid.grid,
        background,
        points.points,
        values,
        variance_ratios,
        pbackground,
        structure,
        max_locations,
    )
    field = np.asarray(field)
    if minvalue is not None:
        field[field < minvalue] = minvalue
    if maxvalue is not None:
        field[field > maxvalue] = maxvalue
    if only_diff:
        field[field == background] = np.nan
    return np.transpose(field)



"""Input data for surfex binaries."""
import json
import logging
import os
import subprocess
from abc import ABC, abstractmethod

"""
def binary_input_data():
    return {
        "pgd": {
            "NAM_FRAC#LECOSG": {
                "True": {
                    "macros": {
                        "VTYPE": {"fmt": "02d"},
                        "DECADE": {"ntime": "NAM_DATA_ISBA#NTIME"},
                    },
                    "NAM_DATA_ISBA#CFTYP_ALBNIR_SOIL": {
                        "DIRTYP": {
                            "NAM_DATA_ISBA#CFNAM_ALBNIR_SOIL": "@ecoclimap_sg@/ALB_@VTYPE@_@DECADE@"
                        }
                    },
                },
                "False": {
                    "ecoclimapI_covers_param.bin": "@ecoclimap_bin_dir@/ecoclimapI_covers_param.bin",
                    "ecoclimapII_eu_covers_param.bin": "@ecoclimap_bin_dir@/ecoclimapII_eu_covers_param.bin",
                    "ecoclimapII_af_covers_param.bin": "@ecoclimap_bin_dir@/ecoclimapII_af_covers_param.bin",
                },
            },
            "NAM_ZS#YFILETYPE": {"DIRECT": {"NAM_ZS#YFILE": "@gmted@/gmted2010"}},
        },
        "prep": {},
    }
"""


class InputDataToSurfexBinaries(ABC):
    """Abstract input data."""

    @abstractmethod
    def __init__(self):
        """Construct."""
        raise NotImplementedError

    @abstractmethod
    def prepare_input(self):
        """Prepare input."""
        raise NotImplementedError


class OutputDataFromSurfexBinaries(ABC):
    """Abstract output data."""

    @abstractmethod
    def __init__(self):
        """Construct."""
        raise NotImplementedError

    @abstractmethod
    def archive_files(self):
        """Archive files."""
        raise NotImplementedError


class JsonOutputData(OutputDataFromSurfexBinaries):
    """Output data."""

    def __init__(self, data):
        """Output data from dict.

        Args:
            data (dict): Output data.

        """
        self.data = data

    def archive_files(self):
        """Archive files."""
        for output_file, target in self.data.items():

            logging.info("%s -> %s", output_file, target)
            command = "mv"
            if isinstance(target, dict):
                for key in target:
                    logging.debug("%s %s %s", output_file, key, target[key])
                    command = target[key]
                    target = key

            cmd = command + " " + output_file + " " + target
            try:
                logging.info(cmd)
                subprocess.check_call(cmd, shell=True)  # noqaS602
            except IOError:
                logging.error("%s failed", cmd)
                raise RuntimeError(cmd + " failed") from IOError


class JsonOutputDataFromFile(JsonOutputData):
    """JSON output data."""

    def __init__(self, file):
        """Construct from json file."""
        with open(file, mode="r", encoding="utf-8") as file_handler:
            data = json.load(file_handler)
        JsonOutputData.__init__(self, data)

    def archive_files(self):
        """Archive files."""
        JsonOutputData.archive_files(self)


class JsonInputData(InputDataToSurfexBinaries):
    """JSON input data."""

    def __init__(self, data):
        """Construct input data.

        Args:
            data (dict): Input data.
        """
        self.data = data

    def prepare_input(self):
        """Prepare input."""
        for target, input_file in self.data.items():

            logging.info("%s -> %s", target, input_file)
            logging.debug(os.path.realpath(target))
            command = None
            if isinstance(input_file, dict):
                for key in input_file:
                    logging.debug(key)
                    logging.debug(input_file[key])
                    command = str(input_file[key])
                    input_file = str(key)
                    command = command.replace("@INPUT@", input_file)
                    command = command.replace("@TARGET@", target)

            if os.path.realpath(target) == os.path.realpath(input_file):
                logging.info("Target and input file is the same file")
            else:
                if command is None:
                    cmd = "ln -sf " + input_file + " " + target
                else:
                    cmd = command
                try:
                    logging.info(cmd)
                    subprocess.check_call(cmd, shell=True)  # noqaS602
                except IOError:
                    raise (cmd + " failed") from IOError

    def add_data(self, data):
        """Add data.

        Args:
            data (dict): Data to add
        """
        for key in data:
            value = data[key]
            self.data.update({key: value})


class JsonInputDataFromFile(JsonInputData):
    """JSON input data."""

    def __init__(self, file):
        """Construct JSON input data.

        Args:
            file (str): JSON file name

        """
        with open(file, mode="r", encoding="utf-8") as file_handler:
            data = json.load(file_handler)
        JsonInputData.__init__(self, data)

    def prepare_input(self):
        """Prepare input."""
        JsonInputData.prepare_input(self)


class InputDataFromNamelist(JsonInputData):
    """Binary input data for offline executables."""

    def __init__(self, nml, input_data, program, platform, basetime=None, validtime=None):
        """Construct InputDataFromNamelist.

        Args:
            nml (f90nml.Namelist): Namelist
            input_data (dict): Input data mapping
            program (str): Kind of program
            platform (SystemFilePaths): Platform settings
            basetime (as_datetime, optional): Baseetime. Defaults to None.
            validtime (as_datetime, optional): Validtime. Defaults to None.

        Raises:
            RuntimeError: Program not defined

        """
        self.nml = nml
        self.platform = platform
        self.basetime = basetime
        self.validtime = validtime
        try:
            self.data = input_data[program]
        except KeyError:
            raise RuntimeError(f"Could not find program {program}") from KeyError
        data = self.process_data()
        JsonInputData.__init__(self, data)

    @staticmethod
    def get_nml_value(nml, block, key, indices=None):
        """Get namelist value.

        Args:
            nml (nmlf90.Namelist): Namelist
            block (str): Namelist block
            key (str): Namelist key
            indices (list, optional): Indices to read. Defaults to None.

        Returns:
            setting (any): Namelist setting

        """
        logging.debug("Checking block=%s key=%s", block, key)
        if block in nml:
            if key in nml[block]:
                logging.debug(nml[block][key])
                if indices is not None:
                    logging.debug("indices=%s", indices)
                    try:
                        if len(indices) == 2:
                            val = nml[block][key][indices[1]][indices[0]]
                        else:
                            val = nml[block][key][indices[0]]
                            logging.debug("Found 1D value %s", val)
                            if isinstance(val, list):
                                return None
                    except IndexError:
                        return None
                    except TypeError:
                        return None
                else:
                    val = nml[block][key]

                logging.debug("Found: %s Indices=%s", val, indices)
                return val
        return None

    @staticmethod
    def get_nml_value_from_string(nml, string, sep="#", indices=None):
        """Get namelist value from a string.

        Args:
            nml (nmlf90.Namelist): Namelist
            string (str): Namelist identifier
            sep (str, optional): _description_. Defaults to "#".
            indices (list, optional): Indices to read. Defaults to None.

        Returns:
            setting (any): Namelist setting

        """
        nam_section = string.split(sep)[0]
        nam_key = string.split(sep)[1]
        return InputDataFromNamelist.get_nml_value(
            nml, nam_section, nam_key, indices=indices
        )

    def substitute(self, key, val, macros=None, micro="@", check_parsing=False):
        """Substitute patterns.

        Args:
            key (str): _description_
            val (str): _description_
            macros (dict, optional): Macros. Defaults to None.
            micro (str, optional): Micro character. Defaults to "@".
            check_parsing (bool, optional): Caheck if values were substituted.

        Returns:
            dict: Substituted key=value

        """
        logging.debug(
            "Substitute key=%s and val=%s %s %s", key, val, self.basetime, self.validtime
        )
        pkey = key
        pval = val
        for spath_key, spath_val in self.platform.system_file_paths.items():
            pkey = pkey.replace(f"{micro}{spath_key}{micro}", spath_val)
            pval = pval.replace(f"{micro}{spath_key}{micro}", spath_val)
        if macros is not None:
            for macro_key, macro_val in macros.items():
                pkey = pkey.replace(f"{micro}{macro_key}{micro}", macro_val)
                pval = pval.replace(f"{micro}{macro_key}{micro}", macro_val)

        pkey = self.platform.parse_setting(
            pkey,
            validtime=self.validtime,
            basedtg=self.basetime,
            check_parsing=check_parsing,
        )
        pval = self.platform.parse_setting(
            pval,
            validtime=self.validtime,
            basedtg=self.basetime,
            check_parsing=check_parsing,
        )
        return {pkey: pval}

    def read_macro_setting(self, macro_defs, key, default=None, sep="#"):
        """Read a macro setting.

        Args:
            macro_defs (dict): Macro definition
            key (str): Macro setting to get.
            default (str, optional): Default value. Defaults to None.
            sep (str, optional): Namelist key separator. Defaults to "#".

        Returns:
            setting (any)
        """
        try:
            setting = macro_defs[key]
            if isinstance(setting, str):
                if setting.find(sep) > 0:
                    logging.debug("Read macro setting from namelist %s", setting)
                    setting = self.get_nml_value_from_string(self.nml, setting)
            return setting
        except KeyError:
            return default

    def extend_macro(self, key, val, macros, sep="#"):
        """Extend entries from macro.

        Args:
            key (_type_): _description_
            val (_type_): _description_
            macros (dict): Macros
            sep (str, optional): Namelist key separator. Defaults to "#".

        Raises:
            NotImplementedError: _description_
            NotImplementedError: _description_

        Returns:
            dict: Key, value dictionary
        """
        logging.debug("macros=%s", macros)
        if macros is None:
            return {key: val}

        processed_data = {}
        for macro, macro_types in macros.items():
            loop = {}
            for macro_type, macro_defs in macro_types.items():
                logging.debug("macro_defs=%s", macro_defs)

                if macro_type == "ekfpert":
                    nncvs = self.read_macro_setting(macro_defs, "list", sep=sep)
                    nncvs = nncvs.copy()
                    duplicate = self.read_macro_setting(macro_defs, "duplicate", sep=sep)
                    if duplicate:
                        nncvs += nncvs
                    loop.update({"0": "0"})
                    icounter1 = 1
                    icounter2 = 1
                    for nncv in nncvs:
                        if nncv == 1:
                            loop.update({str(icounter1): str(icounter2)})
                            icounter1 += 1
                        icounter2 += 1
                elif macro_type == "dict":
                    values = self.get_nml_value_from_string(self.nml, macro_defs)
                    counter = 0
                    for key, val in values.items():
                        loop.update({str(key): str(val)})
                        counter += 1

                elif macro_type == "iterator":
                    start = self.read_macro_setting(macro_defs, "start", sep=sep)
                    end = self.read_macro_setting(macro_defs, "end", sep=sep)
                    fmt = self.read_macro_setting(
                        macro_defs, "fmt", sep=sep, default=None
                    )
                    if fmt is None:
                        fmt = "{:d}"
                    for lval in range(start, end):
                        lval = fmt.format(lval)
                        loop.update({str(lval): str(lval)})
                else:
                    raise NotImplementedError

            # Loop normal macros not being nml arrays
            unprocessed_data = processed_data.copy()
            if processed_data:
                unprocessed_data = processed_data
            else:
                unprocessed_data = {key: val}

            for key, val in unprocessed_data.items():
                for vmacro1, vmacro2 in loop.items():
                    logging.debug(
                        "key=%s val=%s macro=%s vmacro1=%s vmacro2=%s",
                        key,
                        val,
                        macro,
                        vmacro1,
                        vmacro2,
                    )
                    if key.find("#") > 0:
                        key = self.get_nml_value_from_string(self.nml, key, sep=sep)
                    pkey = key.replace(f"@{macro}@", vmacro1)
                    pval = val.replace(f"@{macro}@", vmacro2)
                    processed_data.update({pkey: pval})

        logging.debug("Processed data=%s", processed_data)
        return processed_data

    def process_macro(self, key, val, macros, sep="#", indices=None):
        """Process macro.

        Args:
            key (_type_): _description_
            val (_type_): _description_
            macros (dict): Macros
            sep (str, optional): Namelist key separator. Defaults to "#".
            indices (list, optional): Process macro from namelist indices.

        Raises:
            NotImplementedError: _description_
            NotImplementedError: _description_

        Returns:
            dict: Key, value dictionary
        """
        logging.debug("macros=%s", macros)
        if macros is None:
            return {key: val}

        logging.debug("indices=%s", indices)
        if indices is None:
            return {key: val}

        pkey = key
        pval = val
        for macro in macros:
            lindex = None
            if len(indices) == 2:
                if macro == "DECADE":
                    lindex = indices[1]
                else:
                    lindex = indices[0]
            elif len(indices) == 1:
                lindex = indices[0]
            elif len(indices) > 2:
                raise NotImplementedError

            vmacro = None
            if lindex is not None:
                try:
                    macro_defs = macros[macro]
                    logging.debug("macro_defs=%s", macro_defs)
                except KeyError:
                    logging.warning(
                        "Macro %s not defined. Use index value %s", macro, lindex
                    )
                    vmacro = str(lindex + 1)

                if macro == "VTYPE":
                    vmacro = str(lindex + 1)
                elif "DECADE" in macros:
                    ntime = self.read_macro_setting(macro_defs, "ntime", sep=sep)
                    dec_days = int(360 / float(ntime))
                    dec_start = int(dec_days / 2)
                    dec_end = 360 + dec_start
                    dec = 0
                    for day in range(dec_start, dec_end, dec_days):
                        logging.debug("day=%s, dec=%s lindex=%s", day, dec, lindex)
                        month = int(day / 30) + 1
                        mday = int(day % 30)
                        if dec == lindex:
                            vmacro = f"{month:02d}{mday:02d}"
                        dec += 1

                logging.debug("Substitute @%s@ with %s", macro, vmacro)
                pkey = pkey.replace(f"@{macro}@", vmacro)
                pval = pval.replace(f"@{macro}@", vmacro)
        return {pkey: pval}

    def matching_value(self, data, val, sep="#", indices=None):
        """Match the value. Possibly also read namelist value.

        Args:
            data (dict): Data to check keys for
            val (str): Key to find
            sep (str, optional): Namelist separator. Defaults to "#".
            indices(list, optional): Indices in namelist

        Raises:
            TypeError: Data must be a dict or string

        Returns:
            dict: Matching entry in data.
        """
        if val is not None:
            val = str(val)
        if isinstance(data, dict):
            skeys = list(data.keys())
        elif isinstance(data, str):
            skeys = [data]
        else:
            raise TypeError

        logging.debug("Check for val=%s in skeys=%s", val, skeys)
        for skey in skeys:
            skey_ms = [skey]
            if skey.find(sep) > 0:
                logging.debug("skey=%s is a namelist variable", skey)
                skey_m = self.get_nml_value_from_string(self.nml, skey, indices=indices)

                if isinstance(skey_m, list):
                    skey_ms = skey_m
                else:
                    skey_ms = [skey_m]

            found_data = []
            for skey_m in skey_ms:
                logging.debug("matching val=%s skey=%s skey_m=%s", val, skey, skey_m)
                if val is not None and val == skey_m:
                    logging.debug("Found %s. data=%s", val, data)
                    if skey.find(sep) > 0:
                        if isinstance(data[skey], dict):
                            found_data.append(data[skey][skey_m])
                        elif isinstance(data[skey], str):
                            found_data.append(data[skey])
                        else:
                            found_data.append({skey_m: data[skey]})
                    else:
                        found_data.append(data[skey])

            logging.debug("found_data=%s", found_data)
            if len(found_data) > 0:
                if len(found_data) == 1:
                    return found_data[0]
                return found_data
        logging.warning("Value=%s not found in data", val)
        return None

    def process_data(self, sep="#"):
        """Process input definitions on files to map."""
        data = {}
        logging.debug("Process data: %s", self.data)
        for key, value in self.data.items():
            logging.debug("key=%s", key)
            # Required namelist variable
            if key.find(sep) > 0:
                val = self.get_nml_value_from_string(self.nml, key)
                setting = self.matching_value(value, val, sep=sep)
                if setting is not None:

                    unprocessed_data = {}
                    macros = None
                    if "macros" in setting:
                        macros = setting["macros"]
                    extenders = None
                    if "extenders" in setting:
                        extenders = setting["extenders"]
                    for key1, value1 in setting.items():
                        logging.debug("key1 %s value1 %s", key1, value1)
                        # Macro definition
                        if key1 == "macros" or key1 == "extenders":
                            pass
                        # Extra namelist level
                        elif key1.find(sep) > 0:
                            logging.debug("Key1=%s is a namelist variable", key1)
                            val1 = self.get_nml_value_from_string(self.nml, key1)
                            indices = None
                            if isinstance(val1, list):

                                logging.debug("Got a namelist list variable")
                                indices = []
                                vals1 = []
                                for i, vals in enumerate(val1):
                                    if isinstance(vals, list):
                                        logging.debug(
                                            "i=%s len(vals)=%s val1=%s",
                                            i,
                                            len(vals),
                                            val1,
                                        )
                                        for j in range(0, len(vals)):
                                            ind = [j, i]
                                            vals1.append(val1[i][j])
                                            indices.append(ind)
                                    else:
                                        vals1.append(val1[i])
                                        indices.append([i])
                            else:
                                vals1 = [val1]
                            for ind, val1 in enumerate(vals1):
                                logging.debug(
                                    "ind=%s Checking for val1=%s in setting=%s",
                                    ind,
                                    val1,
                                    setting,
                                )
                                if indices is not None:
                                    lindices = indices[ind]
                                else:
                                    lindices = None
                                setting1 = self.matching_value(
                                    setting, val1, sep=sep, indices=lindices
                                )
                                if setting1 is not None:
                                    logging.debug(
                                        "key1=%s val1=%s Setting1=%s",
                                        key1,
                                        val1,
                                        setting1,
                                    )
                                    if isinstance(setting1, str):
                                        unprocessed_data.update({val1: setting1})
                                    else:
                                        logging.debug(
                                            "setting1=%s, lindices=%s", setting1, lindices
                                        )
                                        for key2, value2 in setting1.items():
                                            if key2.find(sep) > 0:
                                                key2 = self.get_nml_value_from_string(
                                                    self.nml, key2, indices=lindices
                                                )

                                            logging.debug("key2=%s", key2)
                                            sub_macros = self.process_macro(
                                                key2, value2, macros, indices=lindices
                                            )
                                            logging.debug(
                                                "Substituted macros %s", sub_macros
                                            )
                                            unprocessed_data.update(sub_macros)
                                else:
                                    logging.warning(
                                        "Could not match namelist variable %s", key1
                                    )
                        # Direct defintions
                        else:
                            if isinstance(value1, str):
                                logging.debug("key1=%s value1=%s", key1, value1)
                                unprocessed_data.update({key1: value1})
                            else:
                                logging.warning("Did not find a string %s", type(value1))

                    logging.debug(
                        "Manipulate unprocesessed values with general"
                        " or user-defined macros %s",
                        unprocessed_data,
                    )
                    for key, value in unprocessed_data.items():
                        logging.debug("key=%s value=%s", key, value)
                        if isinstance(value, str):
                            if value.endswith(".dir"):
                                hdr_key = key + ".hdr"
                                hdr_val = value.replace(".dir", ".hdr")
                                sub_values = self.substitute(hdr_key, hdr_val)
                                data.update(sub_values)
                                key = key + ".dir"

                        processed_values = self.extend_macro(key, value, extenders)
                        for key, val in processed_values.items():
                            sub_values = self.substitute(key, val)
                            data.update(sub_values)
                else:
                    logging.warning("Could not find the namelist value=%s", val)
            else:
                logging.warning("Expected a namelist variable")
        return data



"""FA support."""
import logging

try:
    from epygram.formats import resource
except ImportError:
    resource = None


from .geo import ConfProj
from .interpolation import Interpolation


class Fa(object):
    """Fichier Arpege."""

    def __init__(self, fname):
        """Construct a FA object.

        Args:
            fname (str): filename
        """
        self.fname = fname
        self.projection = None
        self.lons = None
        self.lats = None
        self.nearest = None
        self.linear = None

    def field(self, varname, validtime):
        """Read a field.

        Args:
            varname (_type_): _description_
            validtime (_type_): _description_

        Raises:
            ModuleNotFoundError: You need epygram to read FA files
            NotImplementedError: Geometry not implemented

        Returns:
            tuple: np.field, surfex.Geometry

        """
        if resource is None:
            raise ModuleNotFoundError("You need epygram to read FA files")
        else:
            fa_file = resource(self.fname, openmode="r")
            field = fa_file.readfield(varname)
            # TODO this might not work with forcing...
            zone = "CI"
            crnrs = field.geometry.gimme_corners_ij(subzone=zone)

            range_x = slice(crnrs["ll"][0], crnrs["lr"][0] + 1)
            range_y = slice(crnrs["lr"][1], crnrs["ur"][1] + 1)
            # TODO: check time
            logging.info(
                "Not checking validtime for FA variable at the moment: %s", str(validtime)
            )

            if (
                field.geometry.name == "lambert"
                or field.geometry.name == "polar_stereographic"
            ):
                n_y = field.geometry.dimensions["Y_CIzone"]
                n_x = field.geometry.dimensions["X_CIzone"]
                lon0 = field.geometry.projection["reference_lon"].get("degrees")
                lat0 = field.geometry.projection["reference_lat"].get("degrees")
                c0, c1 = field.geometry.getcenter()
                lonc = c0.get("degrees")
                latc = c1.get("degrees")
                d_x = field.geometry.grid["X_resolution"]
                d_y = field.geometry.grid["Y_resolution"]
                ilone = field.geometry.dimensions["X"] - n_x
                ilate = field.geometry.dimensions["Y"] - n_y

                domain = {
                    "nam_conf_proj": {"xlon0": lon0, "xlat0": lat0},
                    "nam_conf_proj_grid": {
                        "xloncen": lonc,
                        "xlatcen": latc,
                        "nimax": n_x,
                        "njmax": n_y,
                        "xdx": d_x,
                        "xdy": d_y,
                        "ilone": ilone,
                        "ilate": ilate,
                    },
                }
                geo_out = ConfProj(domain)
                if field.geometry.name == "polar_stereographic":
                    data = field.data[range_y, range_x].T
                else:
                    data = field.data.T
            else:
                raise NotImplementedError(field.geometry.name + " not implemented yet!")
            return data, geo_out

    def points(self, varname, geo, validtime=None, interpolation="nearest"):
        """Read a 2-D field and interpolates it to requested positions.

        Args:
            varname (str): Variable name
            geo (surfex.geo.Geo): Geometry
            validtime (as_datetime): Validtime
            interpolation (str, optional): Interpoaltion method. Defaults to "nearest".

        Returns:
            tuple: field, interpolator

        """
        field, geo_in = self.field(varname, validtime)
        interpolator = Interpolation(interpolation, geo_in, geo)

        field = interpolator.interpolate(field)
        return field, interpolator



"""Variable."""
import copy
import logging

import numpy as np

from .datetime_utils import as_timedelta
from .fa import Fa
from .file import SurfexFileVariable, get_surfex_io_object
from .geo import get_geo_object
from .grib import Grib, Grib1Variable, Grib2Variable
from .input_methods import get_datasources
from .netcdf import Netcdf, NetCDFReadVariable
from .util import parse_filepattern


class Variable(object):
    """New combined variable."""

    def __init__(self, var_type, var_dict, initial_basetime, prefer_forecast=True):
        """Construct variable.

        Args:
            var_type (str): Variable type.
            var_dict (dict): Variable definitions
            initial_basetime (datetime): Initial basetime
            prefer_forecast (bool, optional): Prefer forecasts instead of analysis. Defaults to True.

        Raises:
            NotImplementedError: Variable not implemented
            RuntimeError: No filepattern provided
            RuntimeError: variable must have attribute
            RuntimeError: You can not have larger offset than the frequency of forecasts

        """
        self.var_type = var_type
        if self.var_type == "netcdf":
            mandatory = ["name", "fcint", "offset", "filepattern"]
        elif self.var_type == "grib1":
            mandatory = [
                "parameter",
                "type",
                "level",
                "tri",
                "fcint",
                "offset",
                "filepattern",
            ]
        elif self.var_type == "grib2":
            mandatory = [
                "discipline",
                "parameterCategory",
                "parameterNumber",
                "levelType",
                "level",
                "typeOfStatisticalProcessing",
                "fcint",
                "offset",
                "filepattern",
            ]
        elif self.var_type == "surfex":
            mandatory = ["varname", "fcint", "offset", "filepattern"]
        elif self.var_type == "fa":
            mandatory = ["name", "fcint", "offset", "filepattern"]
        elif self.var_type == "obs":
            mandatory = ["filetype", "fcint", "offset"]
        else:
            raise NotImplementedError

        for mand_val in mandatory:
            if mand_val not in var_dict:
                raise RuntimeError(
                    var_type
                    + " variable must have attribute "
                    + mand_val
                    + " var_dict:"
                    + str(var_dict)
                )

        self.var_dict = copy.deepcopy(var_dict)
        interval = 3600
        if "timestep" in var_dict:
            interval = var_dict["timestep"]
        self.interval = interval
        if "filepattern" in var_dict:
            self.filepattern = var_dict["filepattern"]
        else:
            raise RuntimeError("No filepattern provided")
        self.initial_basetime = initial_basetime
        self.fcint = int(self.var_dict["fcint"])
        self.offset = int(self.var_dict["offset"])
        self.prefer_forecast = prefer_forecast
        if "prefer_forecast" in self.var_dict:
            self.prefer_forecast = self.var_dict["prefer_forecast"]

        if self.offset > self.fcint:
            raise RuntimeError(
                "You can not have larger offset than the frequency of forecasts "
                + str(self.offset)
                + " > "
                + str(self.fcint)
            )

        logging.debug("Constructed variable for %s", str(self.var_dict))

    def get_filename(self, validtime, previoustime=None):
        """Get the filename.

        Args:
            validtime (datetime.datetime): Valid time.
            previoustime (datetime.datetime, optional): Previous valid time. Defaults to None.

        Returns:
            str: Parsed filename

        """
        logging.debug("Set basename for filename")
        basetime = self.get_basetime(validtime, previoustime=previoustime)
        if previoustime is not None:
            validtime = previoustime
        return parse_filepattern(self.filepattern, basetime, validtime)

    def get_filehandler(self, validtime, cache=None, previoustime=None):
        """Get the file handler.

        Args:
            validtime (datetime.datetime): Valid time.
            cache (surfex.Cache, optional): Cache. Defaults to None.
            previoustime (datetime.datetime, optional): Previous valid time. Defaults to None.

        Raises:
            NotImplementedError: Variable type not implemented

        Returns:
            tuple: Filehandler and file name

        """
        filename = self.get_filename(validtime, previoustime=previoustime)
        if cache is not None and cache.file_open(filename):
            file_handler = cache.get_file_handler(filename)
        else:
            if self.var_type == "netcdf":
                file_handler = Netcdf(filename)
            elif self.var_type == "grib1" or self.var_type == "grib2":
                file_handler = Grib(filename)
            elif self.var_type == "fa":
                file_handler = Fa(filename)
            elif self.var_type == "surfex":
                fileformat = None
                if "fileformat" in self.var_dict:
                    fileformat = self.var_dict["fileformat"]
                filetype = None
                if "filetype" in self.var_dict:
                    filetype = self.var_dict["filetype"]
                geo_in = None
                if "geo_input" in self.var_dict:
                    geo_in = self.var_dict["geo_input"]
                elif "geo_input_file" in self.var_dict:
                    geo_in_file = self.var_dict["geo_input_file"]
                    geo_in = get_geo_object(open(geo_in_file, "r", encoding="utf-8"))

                file_handler = get_surfex_io_object(
                    filename, fileformat=fileformat, filetype=filetype, geo=geo_in
                )
            elif self.var_type == "obs":
                var_dict = self.var_dict
                var_dict = {"set": var_dict}
                basetime = self.get_basetime(validtime)
                file_handler = get_datasources(basetime, var_dict)[0]
            else:
                raise NotImplementedError

            if cache is not None:
                cache.set_file_handler(filename, file_handler)

        logging.debug("var_type: %s", self.var_type)
        logging.debug("filename: %s", filename)
        return file_handler, filename

    def read_var_points(self, var, geo, validtime, previoustime=None, cache=None):
        """Read points for a variable.

        Args:
            var (object): Variable object
            geo (surfex.Geo): Surfex geometry
            validtime (datetime.datetime): Valid time
            previoustime (datetime.datetime, optional): Previous valid time for accumulated
                                                        variables. Defaults to None.
            cache (surfex.Cache, optional): Cache. Defaults to None.

        Returns:
            numpy.darray: A numpy array with point values for variable

        """
        interpolation = "bilinear"
        if self.var_type != "obs":
            if "interpolator" in self.var_dict:
                interpolation = self.var_dict["interpolator"]

        logging.debug("set basetime from %s", validtime)

        # TODO put this in a netcdf variable and we don't need this any more
        kwargs = {}
        if self.var_type == "netcdf":
            kwargs.update({"level": None})
            kwargs.update({"units": None})
            if "level" in self.var_dict:
                kwargs.update({"level": [self.var_dict["level"]]})
            if "units" in self.var_dict:
                kwargs.update({"units": str([self.var_dict["units"]][0])})

        filehandler, filename = self.get_filehandler(
            validtime, cache=cache, previoustime=previoustime
        )
        if previoustime is not None:
            validtime = previoustime
            logging.debug("new validtime to read previous time %s", validtime)

        if filehandler is None:
            logging.warning("No file handler exist for this time step")
            field = np.array([len(geo.lons)])
            field = field.fill(np.nan)
        else:
            id_str = None
            if cache is not None:
                id_str = cache.generate_id(self.var_type, var, filename, validtime)

            if cache is not None and cache.is_saved(id_str):
                field = cache.saved_fields[id_str]
                logging.info("Using cached value for %s", id_str)
            else:
                if self.var_type == "obs":
                    __, field, __ = filehandler.points(geo, validtime=validtime)
                else:
                    field, interpolator = filehandler.points(
                        var, geo, interpolation=interpolation, validtime=validtime
                    )

                    field = self.rotate_geographic_wind(field, interpolator)
                if field is not None:
                    logging.debug("field.shape %s", field.shape)

                if cache is not None:
                    logging.debug("ID_STRING %s", id_str)
                    cache.save_field(id_str, field)
        return field

    def set_var(self, validtime=None):
        """Set the variable.

        Args:
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Raises:
            NotImplementedError: Variable not implemented

        Returns:
            tuple: accumulated, instant, var

        """
        accumulated = False
        if self.var_type == "netcdf":
            name = self.var_dict["name"]
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]
            level = None
            if "level" in self.var_dict:
                level = self.var_dict["level"]
                if not isinstance(level, list):
                    level = [level]
            units = None
            if "units" in self.var_dict:
                units = str([self.var_dict["units"]][0])
            member = None
            if "member" in self.var_dict:
                member = self.var_dict["member"]
                if member is not None:
                    if not isinstance(member, list):
                        member = [member]
            var = NetCDFReadVariable(name, level=level, units=units, member=member)
        elif self.var_type == "grib1":
            par = self.var_dict["parameter"]
            typ = self.var_dict["type"]
            level = self.var_dict["level"]
            tri = self.var_dict["tri"]
            if tri == 4:
                accumulated = True
            var = Grib1Variable(par, typ, level, tri)
        elif self.var_type == "grib2":
            discipline = self.var_dict["discipline"]
            p_c = self.var_dict["parameterCategory"]
            p_n = self.var_dict["parameterNumber"]
            l_t = self.var_dict["levelType"]
            lev = self.var_dict["level"]
            tsp = -1
            if "typeOfStatisticalProcessing" in self.var_dict:
                tsp = self.var_dict["typeOfStatisticalProcessing"]
            if tsp == 1:
                accumulated = True
            var = Grib2Variable(discipline, p_c, p_n, l_t, lev, tsp=tsp)
        elif self.var_type == "surfex":
            varname = self.var_dict["varname"]
            layers = None
            if "layers" in self.var_dict:
                layers = self.var_dict["layers"]
            patches = None
            if "patches" in self.var_dict:
                patches = self.var_dict["patches"]
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]
            datatype = "float"
            if "datatype" in self.var_dict:
                datatype = self.var_dict["datatype"]
            tiletype = "FULL"
            if "tiletype" in self.var_dict:
                tiletype = self.var_dict["tiletype"]

            basetime = self.get_basetime(validtime)
            var = SurfexFileVariable(
                varname,
                validtime=validtime,
                patches=patches,
                layers=layers,
                basetime=basetime,
                interval=self.interval,
                datatype=datatype,
                tiletype=tiletype,
            )
        elif self.var_type == "fa":
            var = self.var_dict["name"]
        elif self.var_type == "obs":
            if "varname" in self.var_dict:
                var = self.var_dict["varname"]
            else:
                var = None
        else:
            raise NotImplementedError

        instant = 0
        if accumulated:
            instant = self.interval
        if "instant" in self.var_dict:
            instant = self.var_dict["instant"]
        if "prefer_forecast" in self.var_dict:
            self.prefer_forecast = self.var_dict["prefer_forecast"]

        return accumulated, instant, var

    def read_variable(self, geo, validtime, cache=None):
        """Read the variable.

        Args:
            geo (surfex.Geometry): Geometry to read
            validtime (datetime.datetime): Valid time.
            cache (surfex.cache): Cache. Defaults to None

        Raises:
            RuntimeError: Negative accumulated value found

        Returns:
            np.darray: Field read and interpolated.

        """
        # Set variable info
        accumulated, instant, var = self.set_var(validtime=validtime)
        previous_field = None
        if accumulated:
            # Re-read field
            previoustime = validtime - as_timedelta(seconds=self.interval)
            # Don't read if previous time is older than the very first basetime
            if previoustime >= self.initial_basetime:
                logging.debug("Re-read %s", previoustime)
                previous_field = self.read_var_points(
                    var, geo, validtime=validtime, previoustime=previoustime, cache=cache
                )
            else:
                previous_field = np.zeros([geo.npoints])
        elif instant > 0:
            previous_field = np.zeros([geo.npoints])

        field = self.read_var_points(var, geo, validtime=validtime, cache=cache)

        # Deaccumulate if either two files are read or if instant is > 0.
        if accumulated or instant > 0:
            field = self.deaccumulate(field, previous_field, instant)
            if any(field[field < 0.0]):
                raise RuntimeError("Negative accumulated value found for " + var)
        return field

    def print_variable_info(self):
        """Print variable."""
        logging.debug(":%s:", str(self.var_dict))

    def deaccumulate(self, field, previous_field, instant):
        """Deaccumulate variable.

        Args:
            field (_type_): _description_
            previous_field (_type_): _description_
            instant (_type_): _description_

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_

        """
        logging.debug("Deaccumulate field with: %s", instant)
        if field is None:
            logging.warning("Field is not read properly")
            return None
        else:
            field = np.subtract(field, previous_field)
            if any(field[field < 0.0]):
                neg = []
                for i in range(0, field.shape[0]):
                    if field[i] < 0.0:
                        neg.append(field[i])
                neg = np.asarray(neg)
                logging.warning(
                    "Deaccumulated field has %s negative values. lowest: %s mean: %s",
                    str(neg.shape[0]),
                    str(np.nanmin(neg)),
                    str(np.nanmean(neg)),
                )
            field[field < 0.0] = 0
            if any(field[field < 0.0]):
                raise Exception("Should not be negative values")
            if float(instant) != 0.0:
                field = np.divide(field, float(instant))

            return field

    def get_basetime(self, validtime, previoustime=None, allow_different_basetime=False):
        """Get the basetime of the file.

        Args:
            validtime (datetime.datetime): Valid time
            previoustime (datetime.datetime, optional): Previous valid time. Defaults to None.
            allow_different_basetime (bool, optional): Allow different base times.
                                                       Defaults to False.

        Raises:
            Exception: _description_
            NotImplementedError: _description_

        Returns:
            datetime.datetime: Base time.

        """
        if self.offset < 0:
            raise Exception("Negative offset does not make sense here")

        # Take offset into account
        first = False
        offset = self.offset

        basetime = validtime - as_timedelta(seconds=offset)
        if basetime <= self.initial_basetime:
            first = True
            basetime = self.initial_basetime

        logging.debug("             validtime: %s", validtime)
        logging.debug("          previoustime: %s", previoustime)
        logging.debug("                 first: %s", first)
        logging.debug("                 fcint: %s", self.fcint)
        logging.debug("                offset: %s", self.offset)
        logging.debug("      initial_basetime: %s", self.initial_basetime)
        logging.debug("  Basetime with offset: %s", basetime)

        # Modify based on fcint
        seconds_since_midnight = int(
            (
                basetime - basetime.replace(hour=0, minute=0, second=0, microsecond=0)
            ).total_seconds()
        )
        if seconds_since_midnight == 86400:
            seconds_since_midnight = 0
        basetime_inc = int(
            seconds_since_midnight / int(as_timedelta(seconds=self.fcint).total_seconds())
        )

        prefer_forecast = as_timedelta(seconds=0)
        if seconds_since_midnight == basetime_inc * int(
            as_timedelta(seconds=self.fcint).seconds
        ):
            if first:
                logging.debug("First basetime")
            else:
                if self.prefer_forecast:
                    logging.debug("Prefer forecasts instead of analyis")
                    prefer_forecast = as_timedelta(seconds=self.fcint)
                else:
                    logging.debug("Prefer analysis instead of forecast")

        fcint = as_timedelta(seconds=self.fcint)
        basetime = (
            basetime.replace(hour=0, minute=0, second=0, microsecond=0)
            + (basetime_inc * fcint)
            - prefer_forecast
        )

        if previoustime is not None:
            if allow_different_basetime:
                raise NotImplementedError

        logging.debug("seconds_since_midnight: %s", seconds_since_midnight)
        logging.debug(
            "         cycle seconds: %s",
            basetime_inc * int(as_timedelta(seconds=self.fcint).seconds),
        )
        logging.debug("       prefer_forecast: %s", self.prefer_forecast)
        logging.debug("   prefer_forecast_inc: %s", prefer_forecast)
        logging.debug("          basetime_inc: %s", basetime_inc)
        logging.debug("           Basetime is: %s", basetime)
        return basetime

    def rotate_geographic_wind(self, field, interpolator):
        """Rotate wind.

        Args:
            field (_type_): _description_
            interpolator (_type_): _description_

        Returns:
            _type_: _description_

        """
        rotate_wind = False
        if "rotate_to_geographic" in self.var_dict:
            rotate_wind = self.var_dict["rotate_to_geographic"]
        if rotate_wind:
            field = interpolator.rotate_wind_to_geographic(field)
            return field
        else:
            return field



"""Obsmon."""
import logging

import numpy as np

try:
    import sqlite3
except ImportWarning:
    sqlite3 = None
    logging.warning("Could not import sqlite3 modules")


from .datetime_utils import as_datetime
from .netcdf import read_first_guess_netcdf_file
from .obs import Observation
from .titan import Departure, dataset_from_file


def open_db(dbname):
    """Open database.

    Args:
        dbname (str): File name.

    Raises:
        RuntimeError: Need SQLite

    Returns:
        sqlite3.connect: A connection

    """
    if sqlite3 is None:
        raise RuntimeError("You need SQLITE for obsmon")

    conn = sqlite3.connect(dbname)
    return conn


def close_db(conn):
    """Close data base connection.

    Args:
        conn (sqlite3.connect): Data base connection.
    """
    conn.close()


def create_db(conn, modes, stat_cols):
    """Create data base.

    Args:
        conn (sqlite3.connect): Data base connection.
        modes (_type_): _description_
        stat_cols (_type_): _description_

    """
    cursor = conn.cursor()

    # Create usage table
    cmd = (
        "CREATE TABLE IF NOT EXISTS usage (DTG INT, obnumber INT, obname CHAR(20), "
        "satname CHAR(20), varname CHAR(20), level INT, latitude FLOAT, longitude FLOAT, "
        "statid CHAR(20), obsvalue FLOAT, fg_dep FLOAT, an_dep FLOAT, biascrl FLOAT, "
        "active INT, rejected INT, passive INT, blacklisted INT, anflag INT)"
    )

    cursor.execute(cmd)

    # Create obsmon table
    cmd = (
        "CREATE TABLE IF NOT EXISTS obsmon (DTG INT, obnumber INT, obname CHAR(20), "
        "satname CHAR(20), varname CHAR(20), level INT, passive INT"
    )
    for mode in modes:
        for col in stat_cols:
            cmd = cmd + "," + col + "_" + mode + " FLOAT"

    cmd = cmd + ")"

    cursor.execute(cmd)
    cursor.execute(
        """CREATE INDEX IF NOT EXISTS obsmon_index on usage(DTG,obnumber,obname)"""
    )

    # Save (commit) the changes
    conn.commit()


def populate_usage_db(conn, dtg, varname, observations):
    """Populate usage.

    Args:
        conn (_type_): _description_
        dtg (_type_): _description_
        varname (_type_): _description_
        observations (_type_): _description_

    """
    logging.info("Update usage")
    obnumber = "1"
    obname = "synop"
    satname = "undef"
    level = "0"

    cursor = conn.cursor()
    # Insert a row of data

    def obs2vectors(my_obs):
        return (
            my_obs.lons,
            my_obs.lats,
            my_obs.stids,
            my_obs.elevs,
            my_obs.values,
            my_obs.flags,
            my_obs.fg_dep,
            my_obs.an_dep,
        )

    vectors = np.vectorize(obs2vectors)
    lons, lats, stids, __, values, flags, fg_deps, an_deps = vectors(observations)

    for i, lon_val in enumerate(lons):
        lon = Observation.format_lon(lon_val)
        lat = Observation.format_lat(lats[i])
        stid = str(stids[i])
        if stid == "NA":
            stid = "NULL"
        value = str(values[i])
        if value == "nan":
            value = "NULL"
        if value == "NULL":
            fg_dep = "NULL"
            an_dep = "NULL"
        else:
            if np.isnan(fg_deps[i]):
                fg_dep = "NULL"
            else:
                fg_dep = str(fg_deps[i])
            if np.isnan(an_deps[i]):
                an_dep = "NULL"
            else:
                an_dep = str(an_deps[i])

        status = str(int(flags[i]))
        if status == "0":
            status = "1"

        cmd = (
            "INSERT INTO usage VALUES("
            + str(dtg)
            + ","
            + obnumber
            + ',"'
            + obname
            + '","'
            + satname
            + '","'
            + varname
            + '",'
            + level
            + ","
            + lat
            + ","
            + lon
            + ","
            + stid
            + ","
            + value
            + ","
            + fg_dep
            + ","
            + an_dep
            + ",0,0,0,0,0,"
            + status
            + ")"
        )
        logging.info(cmd)
        cursor.execute(cmd)

    # Save (commit) the changes
    conn.commit()
    logging.info("Updated usage")


def rmse(predictions, targets):
    """Root mean square error.

    Args:
        predictions (_type_): _description_
        targets (_type_): _description_

    Returns:
        _type_: _description_

    """
    if len(predictions) > 0:
        return np.sqrt(np.nanmin(((predictions - targets) ** 2)))
    else:
        return "NULL"


def bias(predictions):
    """Bias.

    Args:
        predictions (_type_): _description_

    Returns:
        _type_: _description_

    """
    if len(predictions) > 0:
        return np.nanmean(np.subtract(predictions, np.nanmean(predictions)))
    else:
        return "NULL"


def absbias(predictions):
    """Absolute bias.

    Args:
        predictions (_type_): _description_

    Returns:
        _type_: _description_

    """
    if len(predictions) > 0:
        return np.nanmean(np.subtract(abs(predictions), np.nanmean(predictions)))
    else:
        return "NULL"


def mean(predictions):
    """Mean.

    Args:
        predictions (_type_): _description_

    Returns:
        _type_: _description_

    """
    if len(predictions) > 0:
        return np.nanmean(predictions)
    else:
        return "NULL"


def calculate_statistics(observations, modes, stat_cols):
    """Statistics.

    Args:
        observations (_type_): _description_
        modes (_type_): _description_
        stat_cols (_type_): _description_

    Raises:
        NotImplementedError: _description_

    Returns:
        _type_: _description_

    """
    values = []
    for i in range(0, len(observations.flags)):
        values.append(observations.values[i])

    statistics = {}
    for mode in modes:
        fg_dep = []
        an_dep = []
        obs = []
        for i, value in enumerate(values):
            land = observations.lafs[i]
            use = False
            if observations.flags[i] == 0:
                use = True
            if mode == "total":
                use = True
            if mode == "land" and land == 1:
                use = True
            if mode == "sea" and land == 0:
                use = True
            if use:
                obs.append(value)
                if not np.isnan(observations.fg_dep[i]):
                    fg_dep.append(observations.fg_dep[i])
                else:
                    fg_dep.append(np.nan)
                if not np.isnan(observations.an_dep[i]):
                    an_dep.append(observations.an_dep[i])
                else:
                    an_dep.append(np.nan)

        fg_dep = np.asarray(fg_dep)
        an_dep = np.asarray(an_dep)
        obs = np.asarray(obs)

        for col in stat_cols:
            tab = col + "_" + mode
            if col == "nobs":
                statistics.update({tab: len(obs)})
            elif col == "fg_bias":
                statistics.update({tab: bias(fg_dep)})
            elif col == "fg_abs_bias":
                statistics.update({tab: absbias(fg_dep)})
            elif col == "fg_rms":
                statistics.update({tab: rmse(np.add(fg_dep, obs), obs)})
            elif col == "fg_dep":
                statistics.update({tab: mean(fg_dep)})
            elif col == "fg_uncorr":
                statistics.update({tab: mean(fg_dep)})
            elif col == "bc":
                statistics.update({tab: 0})
            elif col == "an_bias":
                statistics.update({tab: bias(an_dep)})
            elif col == "an_abs_bias":
                statistics.update({tab: absbias(an_dep)})
            elif col == "an_rms":
                statistics.update({tab: rmse(np.add(an_dep, obs), obs)})
            elif col == "an_dep":
                statistics.update({tab: mean(an_dep)})
            else:
                raise NotImplementedError("Not defined " + col)
    return statistics


def populate_obsmon_db(conn, dtg, statistics, modes, stat_cols, varname):
    """Populate obsmon.

    Args:
        conn (_type_): _description_
        dtg (_type_): _description_
        statistics (_type_): _description_
        modes (_type_): _description_
        stat_cols (_type_): _description_
        varname (_type_): _description_

    Raises:
        Exception: _description_

    """
    logging.info("Update obsmon table")
    obnumber = "1"
    obname = "synop"
    satname = "undef"
    level = "0"

    cursor = conn.cursor()
    cmd = (
        "SELECT * FROM obsmon WHERE DTG=="  # noqa
        + dtg
        + " AND obnumber=="
        + obnumber
        + ' AND obname =="'
        + obname
        + '" AND varname=="'
        + varname
        + '" AND LEVEL == '
        + level
    )  # noqa

    cursor.execute(cmd)
    records = len(cursor.fetchall())
    if records > 1:
        logging.info(cmd)
        raise Exception("You should not have ", records, " in your database")

    if records == 0:
        cmd = (
            "INSERT INTO obsmon VALUES("
            + dtg
            + ","
            + obnumber
            + ',"'
            + obname
            + '","'
            + satname
            + '","'
            + varname
            + '",'
            + level
            + ",0"
        )
    else:
        cmd = "UPDATE obsmon SET "
    first = True
    for mode in modes:
        for col in stat_cols:
            tab = col + "_" + mode
            if records == 0:
                cmd = cmd + "," + str(statistics[tab]) + ""
            else:
                if first:
                    cmd = cmd + "" + tab + "=" + str(statistics[tab])
                else:
                    cmd = cmd + "," + tab + "=" + str(statistics[tab])
            first = False
    if records == 0:
        cmd = cmd + ")"
    else:
        cmd = (
            cmd
            + " WHERE DTG=="
            + dtg
            + " AND obnumber=="
            + obnumber
            + ' AND obname=="'
            + obname
            + '" AND varname=="'
            + varname
            + '" AND LEVEL == '
            + level
        )

    logging.info(cmd)
    cursor.execute(cmd)
    # Save (commit) the changes
    conn.commit()


def write_obsmon_sqlite_file(**kwargs):
    """Write obsmon sqlite file."""
    modes = ["total", "land", "sea"]
    stat_cols = [
        "nobs",
        "fg_bias",
        "fg_abs_bias",
        "fg_rms",
        "fg_dep",
        "fg_uncorr",
        "bc",
        "an_bias",
        "an_abs_bias",
        "an_rms",
        "an_dep",
    ]

    an_time = kwargs["dtg"]
    if isinstance(an_time, str):
        an_time = as_datetime(an_time)
    dtg = an_time.strftime("%Y%m%d%H")
    varname = kwargs["varname"]
    dbname = kwargs["output"]

    operator = "bilinear"
    if "operator" in kwargs:
        operator = kwargs["operator"]

    q_c = kwargs["qc"]
    obs_titan = dataset_from_file(an_time, q_c, skip_flags=[150])

    conn = open_db(dbname)
    create_db(conn, modes, stat_cols)
    fg_file = kwargs["fg_file"]
    fg_var = kwargs["file_var"]
    an_file = kwargs["an_file"]
    an_var = kwargs["file_var"]

    # Only first guess file implemented at the moment
    geo_in, __, an_field, __, __ = read_first_guess_netcdf_file(an_file, an_var)
    geo_in, __, fg_field, __, __ = read_first_guess_netcdf_file(fg_file, fg_var)

    fg_dep = Departure(
        operator, geo_in, obs_titan, fg_field, "first_guess"
    ).get_departure()
    an_dep = Departure(operator, geo_in, obs_titan, an_field, "analysis").get_departure()

    obs_titan = dataset_from_file(
        an_time, q_c, skip_flags=[150, 199], fg_dep=fg_dep, an_dep=an_dep
    )

    populate_usage_db(conn, dtg, varname, obs_titan)
    populate_obsmon_db(
        conn,
        dtg,
        calculate_statistics(obs_titan, modes, stat_cols),
        modes,
        stat_cols,
        varname,
    )
    close_db(conn)



"""Namelist."""
import collections
import logging
import re

import f90nml

from .binary_input import InputDataFromNamelist


class NamelistGenerator(object):
    """Namelist class."""

    def __init__(self, program, config, definitions, assemble=None, consistency=True):
        """Construct a base namelists class.

        Args:
            program (str): Which surfex binary you want to run ["pgd", "prep", "offline", "soda"]
            config (surfex.Configuration): A SURFEX configuration object
            definitions (dict): Namelist definitions
            assemble(dict, optional): Assembly order. Default to None.
            consistency (bool, optional): Check configuration consistency. Defaults to True.
        """
        self.program = program
        self.config = config
        self.nldict = definitions
        macros_defs = {
            "CPGDFILE": "SURFEX#IO#CPGDFILE",
            "CPREPFILE": "SURFEX#IO#CPREPFILE",
            "CSURFFILE": "SURFEX#IO#CSURFFILE",
            "CSURF_FILETYPE": "SURFEX#IO#CSURF_FILETYPE",
            "CFORCING_FILETYPE": "SURFEX#IO#CFORCING_FILETYPE",
            "CTIMESERIES_FILETYPE": "SURFEX#IO#CTIMESERIES_FILETYPE",
            "XRIMAX": "SURFEX#PARAMETERS#XRIMAX",
            "LSPLIT_PATCH": "SURFEX#IO#LSPLIT_PATCH",
            "LFAKETREE": "SURFEX#TREEDRAG#FAKETREES",
            "LECOSG": "SURFEX#COVER#SG",
            "XTSTEP": "SURFEX#IO#XTSTEP",
            "XTSTEP_OUTPUT ": "SURFEX#IO#XTSTEP_OUTPUT",
        }
        macros = {}
        for macro, setting in macros_defs.items():
            vmacro = config.get_setting(setting)
            logging.debug("Mapping macro %s = %s", macro, vmacro)
            if vmacro is not None:
                macros.update({macro: vmacro})

        nobstype = 0
        if program == "soda" or program == "offline":
            nnco = self.config.get_setting("SURFEX#ASSIM#OBS#NNCO")
            nobstype = 0
            for __, obs_val in enumerate(nnco):
                if obs_val == 1:
                    nobstype += 1
            self.config.update_setting("SURFEX#SODA#NOBSTYPE", nobstype)
            soil_assim = self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA")
            nncv = None
            if soil_assim == "EKF":
                nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#EKF#NNCV")
            if soil_assim == "ENKF":
                nncv = self.config.get_setting("SURFEX#ASSIM#ISBA#ENKF#NNCV")
            nvar = 0
            if nncv is not None:
                for __, cval in enumerate(nncv):
                    if cval == 1:
                        nvar += 1
            self.config.update_setting("SURFEX#SODA#NVAR", nvar)

        if program == "soda":
            laesnm = False
            hh = self.config.get_setting("SURFEX#SODA#HH")
            if hh in self.config.get_setting("SURFEX#ASSIM#ISBA#UPDATE_SNOW_CYCLES"):
                laesnm = True
            self.config.update_setting("SURFEX#SODA#LAESNM", laesnm)

        macros = self.flatten_config()
        print(macros)
        self.macros = macros
        if assemble is None:
            self.assemble = self.namelist_blocks()
        else:
            self.assemble = assemble
        logging.debug(self.assemble)
        nlres = self.assemble_namelist()
        self.nml = f90nml.Namelist(nlres)
        if consistency:
            problems = self.concistency(self.nml)
            if problems:
                logging.warning("Found problems!")
                for problem, info in problems.items():
                    for level, desc in info.items():
                        if level == "SEVERE":
                            logging.error(
                                "Problem with: %s Description=%s", problem, desc
                            )
                        elif level == "WARNING":
                            logging.warning(
                                "Problem with: %s Description=%s", problem, desc
                            )
                        else:
                            logging.info("Problem with: %s Description=%s", problem, desc)

    def flatten_config(self):
        """Flatten dictionary.

        Returns:
            source(dict): Flat dict with settings
        """

        def _flatten_dict_gen(dic, parent_key, sep):
            for key, val in dic.items():
                new_key = parent_key + sep + key if parent_key else key
                if isinstance(val, collections.abc.MutableMapping):
                    yield from flatten_dict(val, new_key, sep=sep).items()
                else:
                    yield new_key, val

        def flatten_dict(
            d: collections.abc.MutableMapping, parent_key: str = "", sep: str = "#"
        ):
            return dict(_flatten_dict_gen(d, parent_key, sep))

        return flatten_dict(self.config.settings)

    def namelist_blocks(self):
        """Construct building blocks for the namelist genrator."""
        logging.info("Building namelist blocks for program: %s", self.program)

        input_blocks = ["io", "constants", "treedrag", "flake"]
        lisba = False
        if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
            lisba = True
        elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "3-L":
            lisba = True
        elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "2-L":
            lisba = True

        # Program specific settings
        if self.program == "pgd":

            input_blocks += ["pgd", "pgd_cover", "pgd_zs"]
            eco_sg = self.config.get_setting("SURFEX#COVER#SG")
            if eco_sg:
                input_blocks += ["pgd_ecoclimap_sg"]
            else:
                input_blocks += ["pgd_ecoclimap"]

            # Set ISBA properties
            if lisba:
                input_blocks += ["pgd_isba"]
            if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
                input_blocks += ["pgd_isba_dif"]
            elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "3-L":
                input_blocks += ["pgd_isba_3l"]
            elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "2-L":
                input_blocks += ["pgd_isba_2l"]

            # Set MEB
            if self.config.get_setting("SURFEX#ISBA#MEB"):
                input_blocks += ["pgd_meb"]

            # RSMIN
            if self.config.get_setting("SURFEX#COVER#SG"):
                input_blocks += ["pgd_rsmin_sg"]
            else:
                input_blocks += ["pgd_rsmin"]

            # CV
            if self.config.get_setting("SURFEX#COVER#SG"):
                input_blocks += ["pgd_cv_sg"]
            else:
                input_blocks += ["pgd_cv"]

            # Treedrag
            input_blocks += ["pgd_treedrag"]

            if self.config.get_setting("SURFEX#TOWN#LTOWN_TO_ROCK"):
                input_blocks += ["pgd_town_to_rock"]

            if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
                input_blocks += ["pgd_flake"]

            # Sea
            input_blocks += ["pgd_sea"]

        elif self.program == "prep":
            input_blocks += ["prep"]

            if self.config.get_setting("SURFEX#SEA#ICE") == "SICE":
                input_blocks += ["prep_sice"]

            if self.config.get_setting("SURFEX#TILES#INLAND_WATER") == "FLAKE":
                input_blocks += ["prep_flake"]

            # ISBA
            if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
                input_blocks += ["prep_isba", "prep_isba_dif"]
            elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "3-L":
                input_blocks += ["prep_isba", "prep_isba_3l"]
            elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "2-L":
                input_blocks += ["prep_isba", "prep_isba_2l"]

            if lisba:
                # ISBA CANOPY
                if self.config.get_setting("SURFEX#ISBA#CANOPY"):
                    input_blocks += ["prep_isba_canopy"]

                # Snow
                input_blocks += ["prep_isba_snow"]
                if self.config.get_setting("SURFEX#ISBA#SNOW") == "D95":
                    input_blocks += ["prep_isba_snow_d95"]
                elif self.config.get_setting("SURFEX#ISBA#SNOW") == "3-L":
                    input_blocks += ["prep_isba_snow_3l"]
                if self.config.get_setting("SURFEX#ISBA#SNOW") == "CRO":
                    input_blocks += ["prep_isba_snow_cro"]

            if self.config.get_setting("SURFEX#PREP#FILE") is None:
                input_blocks += ["prep_from_namelist"]
            else:
                input_blocks += ["prep_from_file"]
            if self.config.get_setting("SURFEX#PREP#FILEPGD") is not None:
                input_blocks += ["prep_from_file_with_pgd"]

        elif self.program == "offline" or self.program == "perturbed":
            input_blocks += ["offline"]

            if self.config.get_setting("SURFEX#IO#LSELECT"):
                input_blocks += ["offline_selected_output"]

            # SEAFLX settings
            if self.config.get_setting("SURFEX#TILES#SEA") == "SEAFLX":
                input_blocks += ["offline_seaflux"]

            if self.config.get_setting("SURFEX#SEA#ICE") == "SICE":
                input_blocks += ["offline_sice"]

            # ISBA settings
            if lisba:
                input_blocks += ["offline_isba"]

                if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
                    input_blocks += ["offline_isba_dif"]
                if self.config.get_setting("SURFEX#ISBA#PERTSURF"):
                    input_blocks += ["offline_isba_pertsurf"]

            # SSO
            sso = self.config.get_setting("SURFEX#SSO#SCHEME").lower()
            input_blocks += ["offline_" + sso]

            # Perturbed offline settings
            if lisba:
                input_blocks += ["offline_pert_isba_settings"]
                if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "EKF":
                    input_blocks += ["offline_pert_isba_ekf"]

                if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "ENKF":
                    input_blocks += ["offline_pert_isba_enkf"]

            input_blocks += ["offline_pert_obs"]

            # Climate setting
            if self.config.get_setting("SURFEX#SEA#LVOLATILE_SIC"):
                input_blocks += ["offline_volatile_sic"]

        elif self.program == "soda":
            input_blocks += ["soda", "soda_obs"]

            if self.config.get_setting("SURFEX#SEA#ICE") == "SICE":
                input_blocks += ["soda_sice"]

            # Set OI settings
            if self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "OI":
                input_blocks += ["soda_isba_oi"]

            elif self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "EKF":
                input_blocks += ["soda_isba_ekf"]

            elif self.config.get_setting("SURFEX#ASSIM#SCHEMES#ISBA") == "ENKF":
                input_blocks += ["soda_isba_enkf"]

            # Town
            if self.config.get_setting("SURFEX#ASSIM#SCHEMES#TEB").lower() != "none":
                input_blocks += ["soda_teb"]

        else:
            raise NotImplementedError(self.program)

        return input_blocks

    def get_namelist(self):
        """Get namelist."""
        return self.nml

    @staticmethod
    def check_nml_setting(problems, nml, block, key, value):
        """Check namelist settings.

        Args:
            problems (dict): Problems
            nml (f90nml.Namelist): Namelist
            block (str): Block
            key (str): Key
            value (any): Value

        Returns:
            problems (dict): Problems
        """
        ckey = block + "#" + key
        if block in nml:
            if key in nml[block]:
                if nml[block][key] != value:
                    msg = f"Mismatch: {nml[block][key]} != {value}"
                    problems.update({ckey: {"SEVERE": msg}})
            else:
                problems.update({ckey: {"WARNING": "Namelist key not found"}})
        else:
            problems.update({ckey: {"WARNING": "Namelist block not found"}})
        return problems

    def concistency(self, nml):
        """Check if namelist is consistent with config.

        Args:
            nml (f90nml.Namelist): A parsed f90nml namelist

        Raises:
            NotImplementedError: Mode is not implemented

        Returns:
            problems (dict): Problems.

        """
        logging.info("Check namelist input for program: %s", self.program)

        problems = {}
        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_IO_OFFLINE",
            "CSURF_FILETYPE",
            self.config.get_setting("SURFEX#IO#CSURF_FILETYPE"),
        )
        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_IO_OFFLINE",
            "CTIMESERIES_FILETYPE",
            self.config.get_setting("SURFEX#IO#CTIMESERIES_FILETYPE"),
        )
        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_IO_OFFLINE",
            "CFORCING_FILETYPE",
            self.config.get_setting("SURFEX#IO#CFORCING_FILETYPE"),
        )

        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_IO_OFFLINE",
            "XTSTEP_SURF",
            self.config.get_setting("SURFEX#IO#XTSTEP"),
        )
        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_IO_OFFLINE",
            "XTSTEP_OUTPUT",
            self.config.get_setting("SURFEX#IO#XTSTEP_OUTPUT"),
        )
        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_WRITE_SURF_ATM",
            "LSPLIT_PATCH",
            self.config.get_setting("SURFEX#IO#LSPLIT_PATCH"),
        )

        # Constants and parameters
        problems = self.check_nml_setting(
            problems,
            nml,
            "NAM_SURF_ATM",
            "XRIMAX",
            self.config.get_setting("SURFEX#PARAMETERS#XRIMAX"),
        )

        # Program specific settings
        if self.program == "pgd":

            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_PGD_SCHEMES",
                "CSEA",
                self.config.get_setting("SURFEX#TILES#SEA"),
            )
            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_PGD_SCHEMES",
                "CWATER",
                self.config.get_setting("SURFEX#TILES#INLAND_WATER"),
            )
            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_PGD_SCHEMES",
                "CNATURE",
                self.config.get_setting("SURFEX#TILES#NATURE"),
            )
            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_PGD_SCHEMES",
                "CTOWN",
                self.config.get_setting("SURFEX#TILES#TOWN"),
            )
            if self.config.get_setting("SURFEX#TOWN#LTOWN_TO_ROCK"):

                if self.config.get_setting("SURFEX#TILES#TOWN") != "NONE":
                    logging.warning(
                        "WARNING: TOWN is not NONE and you want LTOWN_TO_ROCK. "
                        "Setting it to NONE!"
                    )

            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_FRAC",
                "LECOSG",
                self.config.get_setting("SURFEX#COVER#SG"),
            )

            fname = str(self.config.get_setting("SURFEX#COVER#YCOVER"))
            __, filetype = self.get_filetype_from_suffix(fname)
            if filetype is not None:
                problems = self.check_nml_setting(
                    problems, nml, "NAM_COVER", "YCOVERFILETYPE", filetype
                )

            # ZS
            fname = str(self.config.get_setting("SURFEX#ZS#YZS"))
            __, filetype = self.get_filetype_from_suffix(fname)
            if filetype is not None:
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ZS", "YZSFILETYPE", filetype
                )

            # Set ISBA properties
            if self.config.get_setting("SURFEX#ISBA#SCHEME") == "DIF":
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ISBA", "CISBA", "DIF"
                )
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ISBA", "NGROUND_LAYER", 14
                )
            elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "3-L":
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ISBA", "CISBA", "3-L"
                )
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ISBA", "NGROUND_LAYER", 3
                )
            elif self.config.get_setting("SURFEX#ISBA#SCHEME") == "2-L":
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ISBA", "CISBA", "2-L"
                )
                problems = self.check_nml_setting(
                    problems, nml, "NAM_ISBA", "NGROUND_LAYER", 2
                )

            # Set patches
            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_ISBA",
                "NPATCH",
                self.config.get_setting("SURFEX#ISBA#NPATCH"),
            )

            # Set MEB
            problems = self.check_nml_setting(
                problems,
                nml,
                "NAM_ISBA",
                "LMEB",
                self.config.get_setting("SURFEX#ISBA#MEB"),
            )

        elif self.program == "prep":
            pass
        elif self.program == "offline" or self.program == "perturbed":
            pass
        elif self.program == "soda":
            pass
        else:
            raise NotImplementedError(self.program)

        return problems

    @staticmethod
    def get_filetype_from_suffix(fname):
        """Get the file type from suffix.

        Args:
            fname (str): File name

        Returns:
            tuple: fname, format

        """
        if fname.endswith(".dir"):
            fname = fname.replace(".dir", "")
            return fname, "DIRECT"
        if fname.endswith(".ascllv"):
            fname = fname.replace(".ascllv", "")
            return fname, "ASCLLV"
        return fname, None

    def input_data_from_namelist(
        self, input_data, platform, basetime=None, validtime=None
    ):
        """Construct a base namelists class to be implemented by namelist implementations.

        Args:
            input_data (dict): Input data definitions
            platform (SystemFilePaths): Platform specific settings
            basetime (as_datetime, optional): Base time
            validtime (as_datetime, optional): Valid time

        Returns:
            data_obj (InputDataFromNamelist): Input data from namelist

        """
        logging.info("Set input data from namelist for program: %s", self.program)

        data_obj = InputDataFromNamelist(
            self.nml,
            input_data,
            self.program,
            platform,
            basetime=basetime,
            validtime=validtime,
        )
        return data_obj

    def assemble_namelist(self):
        """Generate the namelists for 'target'.

        Raises:
            KeyError: Key not found

        Returns:
            nlres (dict): Assembled namelist

        """
        # Read namelist file with all the categories

        # Check target is valid
        cndict = self.assemble
        nldict = self.nldict

        # Start with empty result dictionary
        nlres = {}

        # Assemble the target namelists based on the given category order
        for item in self.flatten_list(cndict):
            catg = item
            # variable substitution removed at this level (may be resurrected)
            # assemble namelists for this category
            if catg in nldict:
                for nl in nldict[catg]:
                    if nl not in nlres:
                        # create the result namelist dict
                        nlres[nl] = {}
                    if catg == "rm{" + nl + "}":
                        # clear/remove the given namelist (but not used for now)
                        nlres[nl].clear()
                    else:
                        for key in nldict[catg][nl]:
                            val = nldict[catg][nl][key]
                            finval = val
                            # Replace ${var-def} with value from config, possibly macro-expanded
                            # For now assumes only one subst. per line, could be generalized if needed
                            if str(finval).find("$") >= 0:
                                m = re.search(
                                    r"^([^\$]*)\$\{([\w\#]+)\-?([^}]*)\}(.*)", str(val)
                                )
                                if m:
                                    pre = m.group(1)
                                    nam = m.group(2)
                                    defval = m.group(3)
                                    post = m.group(4)
                                    logging.debug("macros=%s", self.macros)
                                    logging.debug("look for nam=%s", nam)
                                    try:
                                        repval = self.macros[nam]
                                    except KeyError:
                                        repval = None
                                    if repval is None:
                                        if defval != "":
                                            logging.debug(
                                                "Using default value %s for '%s'",
                                                defval,
                                                nam,
                                            )
                                            repval = self.find_num(defval)
                                        else:
                                            logging.debug("No value found for: '%s'", nam)
                                    else:
                                        logging.debug(
                                            "Replaced %s with: %s", nam, str(repval)
                                        )
                                    if isinstance(repval, str):
                                        finval = str(pre) + str(repval) + str(post)
                                    else:
                                        finval = repval
                                else:
                                    raise KeyError(val)
                            nlres[nl][key] = finval
        return nlres

    def write(self, output_file):
        """Generate the namelists for 'target'.

        Args:
            output_file (str): where to write the result (OPTIONS.nam, fort.4 or EXSEG1.nam typically)

        """
        self.nml.uppercase = True
        self.nml.true_repr = ".TRUE."
        self.nml.false_repr = ".FALSE."
        self.nml.write(output_file, force=True)
        logging.debug("Wrote: %s", output_file)

    @staticmethod
    def flatten_list(li):
        """Recursively flatten a list of lists (of lists)."""
        if li == []:
            return li
        if isinstance(li[0], list):
            return NamelistGenerator.flatten_list(li[0]) + NamelistGenerator.flatten_list(
                li[1:]
            )
        return li[:1] + NamelistGenerator.flatten_list(li[1:])

    @staticmethod
    def find_num(s):
        """Purpose: un-quote numbers."""
        try:
            i = int(s)
            return i
        except ValueError:
            pass
        try:
            f = float(s)
            return f
        except ValueError:
            return s
